#   Copyright 2013 David Malcolm <dmalcolm@redhat.com>
#   Copyright 2013 Red Hat, Inc.
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
#   USA

# Python module for working with Firehose XML files, also, potentially,
# a command-line tool

from collections import OrderedDict, namedtuple
from subprocess import Popen, PIPE
import xml.etree.ElementTree as ET
import hashlib
import glob
import sys
import os

from six import BytesIO, string_types, integer_types, iteritems

_string_type = string_types[0]


class Attribute(namedtuple('Attribute', ('name', 'type', 'nullable'))):
    """
    Description of an attribute of a class.

    name : the name of the attribute

    type : one of:
             * a string containing the name of the type (so that we can make
               forward-references to types)
             * int, float, _string_type, meaning this attribute is of that
               type
             * a list containing one string, meaning this attribute is a
               list of that (named) type

    nullable: boolean: can this attribute be None?
    """
    def __new__(self, name, type, nullable=False):
        return super(Attribute, self).__new__(self, name, type, nullable)

    def resolve_type(self):
        return globals()[self.type]

    def from_json(self, jsonobj):
        if jsonobj is None:
            return jsonobj
        if isinstance(self.type, list):
            # expect a 1-length string containing the internal type of the list
            innertypename = self.type[0]
            innertype = globals()[innertypename]
            return [innertype.from_json(jsonitem) for jsonitem in jsonobj]
        if self.type == _string_type:
            return jsonobj
        if self.type == int:
            return jsonobj
        if self.type == float:
            return jsonobj
        return self.resolve_type().from_json(jsonobj)

def to_json(obj):
    if hasattr(obj, 'to_json'):
        return obj.to_json()
    if isinstance(obj, list):
        return [to_json(item) for item in obj]
    if isinstance(obj, (int, _string_type, float, type(None))):
        return obj
    raise TypeError("don't know how to convert %r to JSON" % obj)

def from_json_using_attrs(cls, jsonobj):
    """
    Given a class cls and a jsonobj, construct an instance of cls, using
    its attrs metadata.
    """
    if jsonobj is None:
        return None
    kwargs = {}
    for attr in cls.attrs:
        attrjson = jsonobj[attr.name]
        kwargs[attr.name] = attr.from_json(attrjson)
    result = cls(**kwargs)
    return result

class JsonMixin(object):
    def to_json(self):
        result = {}
        for attr in self.attrs:
            result[attr.name] = to_json(getattr(self, attr.name))
        return result

    @classmethod
    def from_json(cls, jsonobj):
        return from_json_using_attrs(cls, jsonobj)

    def __eq__(self, other):
        for attr in self.attrs:
            try:
                if getattr(self, attr.name) != getattr(other, attr.name):
                    return False
            except AttributeError:
                return False

        return True

    def __ne__(self, other):
        return not (self == other)

class Analysis(JsonMixin):
    attrs = [Attribute('metadata', 'Metadata'),
             Attribute('results', ['Result']),
             Attribute('customfields', 'CustomFields', nullable=True)]

    def __init__(self, metadata, results, customfields=None):
        assert isinstance(metadata, Metadata)
        assert isinstance(results, list)
        if customfields is not None:
            assert isinstance(customfields, CustomFields)
        for result in results:
            assert isinstance(result, Result)

        self.metadata = metadata
        self.results = results
        self.customfields = customfields

    @classmethod
    def from_xml(cls, fileobj):
        tree = ET.parse(fileobj)
        root = tree.getroot()

        metadata = Metadata.from_xml(root.find('metadata'))
        results_node = root.find('results')
        results = []
        for result_node in results_node:
            if result_node.tag == 'issue':
                results.append(Issue.from_xml(result_node))
            elif result_node.tag == 'failure':
                results.append(Failure.from_xml(result_node))
            elif result_node.tag == 'info':
                results.append(Info.from_xml(result_node))
        customfields_node = root.find('custom-fields')
        if customfields_node is not None:
            customfields = CustomFields.from_xml(customfields_node)
        else:
            customfields = None
        return Analysis(metadata, results, customfields)

    def to_xml(self):
        tree = ET.ElementTree()
        node = ET.Element('analysis')
        tree._setroot(node)
        node.append(self.metadata.to_xml())
        results_node = ET.Element('results')
        node.append(results_node)
        for result in self.results:
            results_node.append(result.to_xml())
        if self.customfields is not None:
            node.append(self.customfields.to_xml())
        return tree

    def to_xml_bytes(self):
        xml = self.to_xml()
        output = BytesIO()
        xml.write(output, encoding='utf-8')
        return output.getvalue()

    def __repr__(self):
        return ('Analysis(metadata=%r, results=%r, customfields=%r)'
                % (self.metadata, self.results, self.customfields))

    def __hash__(self):
        # (self.results is a list and is thus not hashable)
        return hash(self.metadata)

    def accept(self, visitor):
        visitor.visit_analysis(self)
        self.metadata.accept(visitor)
        for result in self.results:
            result.accept(visitor)

    def fixup_files(self, relativedir=None, hashalg=None):
        """
        Record the absolute path of each file, and record the digest of the
        file content
        """
        class FixupFiles(Visitor):
            def __init__(self, relativedir, hashalg):
                self.relativedir = relativedir
                self.hashalg = hashalg

            def visit_file(self, file_):
                if self.relativedir is not None:
                    file_.abspath = os.path.normpath(os.path.join(self.relativedir,
                                                                  file_.givenpath))

                if hashalg is not None:
                    bestpath = file_.abspath \
                        if file_.abspath else file_.givenpath

                    with open(bestpath, 'rb') as f:
                        h = hashlib.new(hashalg)
                        h.update(f.read())
                        file_.hash_ = Hash(alg=hashalg, hexdigest=h.hexdigest())

        visitor = FixupFiles(relativedir, hashalg)
        self.accept(visitor)

    def set_custom_field(self, name, value):
        if self.customfields is None:
            self.customfields = CustomFields()
        self.customfields[name] = value

class Result(JsonMixin):
    @classmethod
    def from_json(cls, jsonobj):
        if jsonobj['type'] == 'Issue':
            return from_json_using_attrs(Issue, jsonobj)
        elif jsonobj['type'] == 'Failure':
            return from_json_using_attrs(Failure, jsonobj)
        elif jsonobj['type'] == 'Info':
            return from_json_using_attrs(Info, jsonobj)
        raise TypeError('unknown type: %r' % jsonobj['type'])

class Issue(Result):
    attrs = [Attribute('cwe', int, nullable=True),
             Attribute('testid', _string_type, nullable=True),
             Attribute('location', 'Location'),
             Attribute('message', 'Message'),
             Attribute('notes', 'Notes', nullable=True),
             Attribute('trace', 'Trace', nullable=True),
             Attribute('severity', _string_type, nullable=True),
             Attribute('customfields', 'CustomFields', nullable=True)]

    def __init__(self,
                 cwe,
                 testid,
                 location,
                 message,
                 notes,
                 trace,
                 severity=None,
                 customfields=None):
        if cwe is not None:
            assert isinstance(cwe, int)
        if testid is not None:
            assert isinstance(testid, _string_type)
        assert isinstance(location, Location)
        assert isinstance(message, Message)
        if notes:
            assert isinstance(notes, Notes)
        if trace:
            assert isinstance(trace, Trace)
        if severity is not None:
            assert isinstance(severity, _string_type)
        if customfields is not None:
            assert isinstance(customfields, CustomFields)
        self.cwe = cwe
        self.testid = testid
        self.location = location
        self.message = message
        self.notes = notes
        self.trace = trace
        self.severity = severity
        self.customfields = customfields

    @classmethod
    def from_xml(cls, node):
        cwe = node.get('cwe')
        if cwe is not None:
            cwe = int(cwe)
        testid = node.get('test-id')
        location = Location.from_xml(node.find('location'))
        message = Message.from_xml(node.find('message'))
        notes_node = node.find('notes')
        if notes_node is not None:
            notes = Notes.from_xml(notes_node)
        else:
            notes = None
        trace_node = node.find('trace')
        if trace_node is not None:
            trace = Trace.from_xml(trace_node)
        else:
            trace = None
        severity = node.get('severity')
        customfields_node = node.find('custom-fields')
        if customfields_node is not None:
            customfields = CustomFields.from_xml(customfields_node)
        else:
            customfields = None
        return Issue(cwe, testid, location, message, notes, trace, severity, customfields)

    def to_xml(self):
        node = ET.Element('issue')
        if self.cwe is not None:
            node.set('cwe', str(self.cwe))
        if self.testid is not None:
            node.set('test-id', str(self.testid))
        node.append(self.message.to_xml())
        if self.notes:
            node.append(self.notes.to_xml())
        node.append(self.location.to_xml())
        if self.trace:
            node.append(self.trace.to_xml())
        if self.severity is not None:
            node.set('severity', str(self.severity))
        if self.customfields is not None:
            node.append(self.customfields.to_xml())
        return node

    def to_json(self):
        jsonobj = JsonMixin.to_json(self)
        jsonobj['type'] = 'Issue'
        return jsonobj

    def write_as_gcc_output(self, out):
        """
        Write the report in the style of a GCC warning to the given
        file-like object
        """
        def writeln(msg):
            out.write('%s\n' % msg)
        def diagnostic(filename, line, column, kind, msg):
            out.write('%s:%i:%i: %s: %s\n'
                      % (filename, line, column,
                         kind, msg))
        if self.location.function is not None:
            writeln("%s: In function '%s':"
                    % (self.location.file.givenpath,
                       self.location.function.name))
        if self.cwe:
            cwetext = ' [%s]' % self.get_cwe_str()
        else:
            cwetext = ''
        diagnostic(filename=self.location.file.givenpath,
                   line=self.location.line,
                   column=self.location.column,
                   kind='warning',
                   msg='%s%s' % (self.message.text, cwetext))
        if self.notes:
            writeln(self.notes.text.rstrip())
        if self.trace:
            for state in self.trace.states:
                notes = state.notes
                diagnostic(filename=state.location.file.givenpath,
                           line=state.location.line,
                           column=state.location.column,
                           kind='note',
                           msg=notes.text if notes else '')

    def __repr__(self):
        return ('Issue(cwe=%r, testid=%r, location=%r, message=%r,'
                ' notes=%r, trace=%r, severity=%r, customfields=%r)'
                % (self.cwe, self.testid, self.location, self.message,
                   self.notes, self.trace, self.severity, self.customfields))

    def __hash__(self):
        return (hash(self.cwe) ^ hash(self.testid)
                ^ hash(self.location) ^ hash(self.message)
                ^ hash(self.notes) ^ hash(self.trace) ^ hash(self.severity))

    def accept(self, visitor):
        visitor.visit_warning(self)
        self.location.accept(visitor)
        self.message.accept(visitor)
        if self.notes:
            self.notes.accept(visitor)
        if self.trace:
            self.trace.accept(visitor)

    def get_cwe_str(self):
        if self.cwe is not None:
            return 'CWE-%i' % self.cwe

    def get_cwe_url(self):
        if self.cwe is not None:
            return 'http://cwe.mitre.org/data/definitions/%i.html' % self.cwe

class Failure(Result):
    attrs = [Attribute('failureid', _string_type, nullable=True),
             Attribute('location', 'Location'),
             Attribute('message', 'Message'),
             Attribute('customfields', 'CustomFields', nullable=True)]

    def __init__(self, failureid, location, message, customfields):
        if failureid is not None:
            assert isinstance(failureid, _string_type)
        if location is not None:
            assert isinstance(location, Location)
        if message is not None:
            assert isinstance(message, Message)
        if customfields is not None:
            assert isinstance(customfields, CustomFields)
        self.failureid = failureid
        self.location = location
        self.message = message
        self.customfields = customfields

    @classmethod
    def from_xml(cls, node):
        failureid = node.get('failure-id')
        location_node = node.find('location')
        if location_node is not None:
            location = Location.from_xml(location_node)
        else:
            location = None
        message_node = node.find('message')
        if message_node is not None:
            message = Message.from_xml(message_node)
        else:
            message = None
        customfields_node = node.find('custom-fields')
        if customfields_node is not None:
            customfields = CustomFields.from_xml(customfields_node)
        else:
            customfields = None
        return Failure(failureid, location, message, customfields)

    def to_xml(self):
        node = ET.Element('failure')

        if self.failureid is not None:
            node.set('failure-id', self.failureid)

        if self.location is not None:
            node.append(self.location.to_xml())

        if self.message is not None:
            node.append(self.message.to_xml())

        if self.customfields is not None:
            node.append(self.customfields.to_xml())

        return node

    def to_json(self):
        jsonobj = JsonMixin.to_json(self)
        jsonobj['type'] = 'Failure'
        return jsonobj

    def __repr__(self):
        return ('Failure(failureid=%r, location=%r, message=%r, customfields=%r)'
                % (self.failureid, self.location, self.message, self.customfields))

    def __hash__(self):
        return (hash(self.failureid) ^ hash(self.location)
                ^ hash(self.message) ^ hash(self.customfields))

    def accept(self, visitor):
        visitor.visit_failure(self)
        if self.location:
            self.location.accept(visitor)
        if self.message:
            self.message.accept(visitor)

class Info(Result):
    attrs = [Attribute('infoid', _string_type, nullable=True),
             Attribute('location', 'Location', nullable=True),
             Attribute('message', 'Message', nullable=True),
             Attribute('customfields', 'CustomFields', nullable=True)]

    def __init__(self, infoid, location, message, customfields):
        if infoid is not None:
            assert isinstance(infoid, _string_type)
        if location is not None:
            assert isinstance(location, Location)
        if message is not None:
            assert isinstance(message, Message)
        if customfields is not None:
            assert isinstance(customfields, CustomFields)
        self.infoid = infoid
        self.location = location
        self.message = message
        self.customfields = customfields

    @classmethod
    def from_xml(cls, node):
        infoid = node.get('info-id')
        location_node = node.find('location')
        if location_node is not None:
            location = Location.from_xml(location_node)
        else:
            location = None
        message_node = node.find('message')
        if message_node is not None:
            message = Message.from_xml(message_node)
        else:
            message = None
        customfields_node = node.find('custom-fields')
        if customfields_node is not None:
            customfields = CustomFields.from_xml(customfields_node)
        else:
            customfields = None
        return Info(infoid, location, message, customfields)

    def to_xml(self):
        node = ET.Element('info')

        if self.infoid is not None:
            node.set('info-id', self.infoid)

        if self.location is not None:
            node.append(self.location.to_xml())

        if self.message is not None:
            node.append(self.message.to_xml())

        if self.customfields is not None:
            node.append(self.customfields.to_xml())

        return node

    def to_json(self):
        jsonobj = JsonMixin.to_json(self)
        jsonobj['type'] = 'Info'
        return jsonobj

    def __repr__(self):
        return ('Info(infoid=%r, location=%r, message=%r, customfields=%r)'
                % (self.infoid, self.location, self.message, self.customfields))

    def __hash__(self):
        return (hash(self.infoid) ^ hash(self.location)
                ^ hash(self.message) ^ hash(self.customfields))

    def accept(self, visitor):
        visitor.visit_info(self)
        if self.location:
            self.location.accept(visitor)
        if self.message:
            self.message.accept(visitor)

class Metadata(JsonMixin):
    attrs = [Attribute('generator', 'Generator'),
             Attribute('sut', 'Sut', nullable=True),
             Attribute('file_', 'File', nullable=True),
             Attribute('stats', 'Stats', nullable=True)]

    def __init__(self, generator, sut, file_, stats):
        assert isinstance(generator, Generator)
        if sut is not None:
            assert isinstance(sut, Sut)
        if file_ is not None:
            assert isinstance(file_, File)
        if stats is not None:
            assert isinstance(stats, Stats)
        self.generator = generator
        self.sut = sut
        self.file_ = file_
        self.stats = stats

    @classmethod
    def from_xml(cls, node):
        generator = Generator.from_xml(node.find('generator'))
        sut_node = node.find('sut')
        if sut_node is not None:
            sut = Sut.from_xml(sut_node)
        else:
            sut = None
        file_node = node.find('file')
        if file_node is not None:
            file_ = File.from_xml(file_node)
        else:
            file_ = None
        stats_node = node.find('stats')
        if stats_node is not None:
            stats = Stats.from_xml(stats_node)
        else:
            stats = None
        result = Metadata(generator, sut, file_, stats)
        return result

    def to_xml(self):
        node = ET.Element('metadata')
        node.append(self.generator.to_xml())
        if self.sut is not None:
            node.append(self.sut.to_xml())
        if self.file_ is not None:
            node.append(self.file_.to_xml())
        if self.stats is not None:
            node.append(self.stats.to_xml())
        return node

    def __repr__(self):
        return ('Metadata(generator=%r, sut=%r, file_=%r, stats=%r)'
                % (self.generator, self.sut, self.file_, self.stats))

    def __hash__(self):
        return (hash(self.generator) ^ hash(self.sut)
                ^ hash(self.file_) ^ hash(self.stats))

    def accept(self, visitor):
        visitor.visit_metadata(self)
        self.generator.accept(visitor)
        if self.sut:
            self.sut.accept(visitor)
        if self.file_:
            self.file_.accept(visitor)
        if self.stats:
            self.stats.accept(visitor)

class Generator(JsonMixin):
    attrs = [Attribute('name', _string_type),
             Attribute('version', _string_type, nullable=True),
             ]

    def __init__(self, name, version=None):
        assert isinstance(name, _string_type)
        if version is not None:
            assert isinstance(version, _string_type)
        self.name = name
        self.version = version

    @classmethod
    def from_xml(cls, node):
        result = Generator(name=node.get('name'),
                           version=node.get('version')) # optional
        return result

    def to_xml(self):
        node = ET.Element('generator')
        node.set('name', self.name)
        if self.version is not None:
            node.set('version', self.version)
        return node

    def __repr__(self):
        return ('Generator(name=%r, version=%r)'
                % (self.name, self.version))

    def __hash__(self):
        return hash(self.name) ^ hash(self.version)

    def accept(self, visitor):
        visitor.visit_generator(self)

class Sut(JsonMixin):
    # FIXME: this part of the schema needs more thought/work

    @classmethod
    def from_xml(cls, node):
        srpm_node = node.find('source-rpm')
        if srpm_node is not None:
            return SourceRpm.from_xml(srpm_node)

        dsc_node = node.find('debian-source')
        if dsc_node is not None:
            return DebianSource.from_xml(dsc_node)

        deb_node = node.find('debian-binary')
        if deb_node is not None:
            return DebianBinary.from_xml(deb_node)

        raise ValueError('unknown sut kind')

    def to_xml(self):
        innernode = self._to_xml_inner_node()
        node = ET.Element('sut')
        node.append(innernode)
        return node

    def _to_xml_inner_node(self):
        raise NotImplementedError

    @classmethod
    def from_json(cls, jsonobj):
        subclsname = jsonobj['type']
        subcls = globals()[subclsname]
        return from_json_using_attrs(subcls, jsonobj)

    def accept(self, visitor):
        visitor.visit_sut(self)

class SourceRpm(Sut):
    attrs = [Attribute('name', _string_type),
             Attribute('version', _string_type),
             Attribute('release', _string_type),
             Attribute('buildarch', _string_type),
             ]

    def __init__(self, name, version, release, buildarch):
        assert isinstance(name, _string_type)
        assert isinstance(version, _string_type)
        assert isinstance(release, _string_type)
        assert isinstance(buildarch, _string_type)
        self.name = name
        self.version = version
        self.release = release
        self.buildarch = buildarch

    @classmethod
    def from_xml(cls, node):
        result = SourceRpm(name=node.get('name'),
                           version=node.get('version'),
                           release=node.get('release'),
                           buildarch=node.get('build-arch'))
        return result

    def _to_xml_inner_node(self):
        node = ET.Element('source-rpm')
        node.set('name', self.name)
        node.set('version', self.version)
        node.set('release', self.release)
        node.set('build-arch', self.buildarch)
        return node

    def to_json(self):
        result = JsonMixin.to_json(self)
        result['type']= 'SourceRpm'
        return result

    def __repr__(self):
        return ('SourceRpm(name=%r, version=%r, release=%r, buildarch=%r)'
                % (self.name, self.version, self.release, self.buildarch))

    def __hash__(self):
        return (hash(self.name) ^ hash(self.version)
                ^ hash(self.release) ^ hash(self.buildarch))


class DebianBinary(Sut):
    """
    Internal Firehose represntation of a Debian binary package. This Object
    is extremely similar to a SourceRpm.
    """
    attrs = [Attribute('name', _string_type),
             Attribute('version', _string_type),
             Attribute('release', _string_type, nullable=True),
             Attribute('buildarch', _string_type)]

    def __init__(self, name, version, release, buildarch):
        """
        Simple constructor. Name should be the *binary* package name,
        version should match Upstream's version number, and release (if
        given) should be the Debian package local version. This should
        only be ommited if the package is a Debian Native package.

        buildarch is the Debian binary arch (like amd64, armhf, hurd-i386)
        """
        assert isinstance(name, _string_type)
        assert isinstance(version, _string_type)
        assert (isinstance(release, _string_type) or release is None)
        assert isinstance(buildarch, _string_type)

        if release is None and "-" in version:
            # XXX: Do we have a better Exception for here?
            raise Exception("Native package with dash in the version string")

        self.name = name
        self.version = version
        self.release = release
        self.buildarch = buildarch

    @classmethod
    def from_xml(cls, node):
        """
        Construct a DebianSource object from an XML payload.
        """
        result = DebianBinary(name=node.get('name'),
                              version=node.get('version'),
                              release=node.get('release'),
                              buildarch=node.get('build-arch'))
        return result

    def _to_xml_inner_node(self):
        """
        (internal use only)

        Produce a DebianSource XML ET for searlizing the data back down to
        XML again.
        """
        node = ET.Element('debian-binary')
        node.set('name', self.name)
        node.set('version', self.version)
        if not self.release is None:
            node.set('release', self.release)
        node.set('build-arch', self.buildarch)
        return node

    def to_json(self):
        result = JsonMixin.to_json(self)
        result['type']= 'DebianBinary'
        return result

    def __repr__(self):
        return ('DebianBinary(name=%r, version=%r, release=%r, arch=%r)'
                % (self.name, self.version, self.release, self.buildarch))

    def __hash__(self):
        return (hash(self.name) ^ hash(self.version)
                ^ hash(self.release) ^ hash(self.buildarch))


class DebianSource(Sut):
    """
    Internal Firehose represntation of a Debian source package. This Object
    is extremely similar to a SourceRpm, but does not include the `buildarch`
    attribute.
    """
    attrs = [Attribute('name', _string_type),
             Attribute('version', _string_type),
             Attribute('release', _string_type, nullable=True)]

    def __init__(self, name, version, release):
        """
        Simple constructor. Name should be the *source* package name,
        version should match Upstream's version number, and release (if
        given) should be the Debian package local version. This should
        only be ommited if the package is a Debian Native package.
        """
        assert isinstance(name, _string_type)
        assert isinstance(version, _string_type)
        assert (isinstance(release, _string_type) or release is None)
        if release is None and "-" in version:
            # XXX: Do we have a better Exception for here?
            raise Exception("Native package with dash in the version string")

        self.name = name
        self.version = version
        self.release = release

    @classmethod
    def from_xml(cls, node):
        """
        Construct a DebianSource object from an XML payload.
        """
        result = DebianSource(name=node.get('name'),
                              version=node.get('version'),
                              release=node.get('release'))
        return result

    def _to_xml_inner_node(self):
        """
        (internal use only)

        Produce a DebianSource XML ET for searlizing the data back down to
        XML again.
        """
        node = ET.Element('debian-source')
        node.set('name', self.name)
        node.set('version', self.version)
        if not self.release is None:
            node.set('release', self.release)
        return node

    def to_json(self):
        result = JsonMixin.to_json(self)
        result['type']= 'DebianSource'
        return result

    def __repr__(self):
        return ('DebianSource(name=%r, version=%r, release=%r)'
                % (self.name, self.version, self.release))

    def __hash__(self):
        return (hash(self.name) ^ hash(self.version)
                ^ hash(self.release))


class Stats(JsonMixin):
    attrs = [Attribute('wallclocktime', float)]

    def __init__(self, wallclocktime):
        assert isinstance(wallclocktime, float)
        self.wallclocktime = wallclocktime

    @classmethod
    def from_xml(cls, node):
        wallclocktime = float(node.get('wall-clock-time'))
        result = Stats(wallclocktime)
        return result

    def to_xml(self):
        node = ET.Element('stats')
        node.set('wall-clock-time', str(self.wallclocktime))
        return node

    def __repr__(self):
        return 'Stats(wallclocktime=%r)' % (self.wallclocktime, )

    def __hash__(self):
        return hash(self.wallclocktime)

    def accept(self, visitor):
        visitor.visit_stats(self)

class Message(JsonMixin):
    attrs = [Attribute('text', _string_type)]

    def __init__(self, text):
        assert isinstance(text, _string_type)
        self.text = text

    @classmethod
    def from_xml(cls, node):
        result = Message(node.text)
        return result

    def to_xml(self):
        node = ET.Element('message')
        node.text = self.text
        return node

    def __repr__(self):
        return 'Message(text=%r)' % (self.text, )

    def __hash__(self):
        return hash(self.text)

    def accept(self, visitor):
        visitor.visit_message(self)

class Notes(JsonMixin):
    attrs = [Attribute('text', _string_type)]

    def __init__(self, text):
        assert isinstance(text, _string_type)
        self.text = text

    @classmethod
    def from_xml(cls, node):
        text = node.text
        result = Notes(text)
        return result

    def to_xml(self):
        node = ET.Element('notes')
        node.text = self.text
        return node

    def __repr__(self):
        return 'Notes(text=%r)' % (self.text, )

    def __hash__(self):
        return hash(self.text)

    def accept(self, visitor):
        visitor.visit_notes(self)

class Trace(JsonMixin):
    attrs = [Attribute('states', ['State'])]

    def __init__(self, states):
        assert isinstance(states, list)
        self.states = states

    def add_state(self, state):
        self.states.append(state)

    @classmethod
    def from_xml(cls, node):
        states = []
        for state_node in node.findall('state'):
            states.append(State.from_xml(state_node))
        result = Trace(states)
        return result

    def to_xml(self):
        node = ET.Element('trace')
        for state in self.states:
            node.append(state.to_xml())
        return node

    def __repr__(self):
        return 'Trace(states=%r)' % (self.states, )

    def __hash__(self):
        result = 0
        for state in self.states:
            result ^= hash(state)
        return result

    def accept(self, visitor):
        visitor.visit_notes(self)
        for state in self.states:
            state.accept(visitor)

class State(JsonMixin):
    attrs = [Attribute('location', 'Location'),
             Attribute('notes', 'Notes', nullable=True)]

    def __init__(self, location, notes):
        assert isinstance(location, Location)
        if notes is not None:
            assert isinstance(notes, Notes)
        self.location = location
        self.notes = notes

    @classmethod
    def from_xml(cls, node):
        location = Location.from_xml(node.find('location'))
        notes_node = node.find('notes')
        if notes_node is not None:
            notes = Notes.from_xml(notes_node)
        else:
            notes = None
        return State(location, notes)

    def to_xml(self):
        node = ET.Element('state')
        node.append(self.location.to_xml())
        if self.notes:
            node.append(self.notes.to_xml())
        return node

    def __repr__(self):
        return 'State(location=%r, notes=%r)' % (self.location, self.notes)

    def __hash__(self):
        return hash(self.location) ^ hash(self.notes)

    def accept(self, visitor):
        visitor.visit_state(self)
        self.location.accept(visitor)
        if self.notes:
            self.notes.accept(visitor)

class Location(JsonMixin):
    attrs = [Attribute('file', 'File'),
             Attribute('function', 'Function', nullable=True),
             Attribute('point', 'Point', nullable=True),
             Attribute('range_', 'Range', nullable=True)]

    def __init__(self, file, function, point=None, range_=None):
        assert isinstance(file, File)
        if function is not None:
            assert isinstance(function, Function)
        if point is not None:
            assert isinstance(point, Point)
        if range_ is not None:
            assert isinstance(range_, Range)
        self.file = file
        self.function = function
        self.point = point
        self.range_ = range_

    @classmethod
    def from_xml(cls, node):
        file = File.from_xml(node.find('file'))
        function_node = node.find('function')
        if function_node is not None:
            function = Function.from_xml(function_node)
        else:
            function = None
        point_node = node.find('point')
        if point_node is not None:
            point = Point.from_xml(point_node)
        else:
            point = None
        range_node = node.find('range')
        if range_node is not None:
            range_ = Range.from_xml(range_node)
        else:
            range_ = None
        return Location(file, function, point, range_)

    def to_xml(self):
        node = ET.Element('location')
        node.append(self.file.to_xml())
        if self.function is not None:
            node.append(self.function.to_xml())
        if self.point is not None:
            node.append(self.point.to_xml())
        if self.range_ is not None:
            node.append(self.range_.to_xml())
        return node

    def __repr__(self):
        return ('Location(file=%r, function=%r, point=%r, range_=%r)' %
                (self.file, self.function, self.point, self.range_))

    def __hash__(self):
        return (hash(self.file) ^ hash(self.function)
                ^ hash(self.point) ^ hash(self.range_))

    def accept(self, visitor):
        visitor.visit_location(self)
        self.file.accept(visitor)
        if self.function:
            self.function.accept(visitor)
        if self.point:
            self.point.accept(visitor)
        if self.range_:
            self.range_.accept(visitor)

    @property
    def line(self):
        if self.point is not None:
            return self.point.line
        if self.range_ is not None:
            return self.range_.start.line

    @property
    def column(self):
        if self.point is not None:
            return self.point.column
        if self.range_ is not None:
            return self.range_.start.column

class File(JsonMixin):
    attrs = [Attribute('givenpath', _string_type),
             Attribute('abspath',  _string_type, nullable=True),
             Attribute('hash_',  'Hash', nullable=True)]

    def __init__(self, givenpath, abspath, hash_=None):
        assert isinstance(givenpath, _string_type)
        if abspath is not None:
            assert isinstance(abspath, _string_type)
        if hash_ is not None:
            assert isinstance(hash_, Hash)

        self.givenpath = givenpath
        self.abspath = abspath
        self.hash_ = hash_

    @classmethod
    def from_xml(cls, node):
        givenpath = node.get('given-path')
        abspath = node.get('absolute-path')
        hash_node = node.find('hash')
        if hash_node is not None:
            hash_ = Hash.from_xml(hash_node)
        else:
            hash_ = None
        result = File(givenpath, abspath, hash_)
        return result

    def to_xml(self):
        node = ET.Element('file')
        node.set('given-path', self.givenpath)
        if self.abspath:
            node.set('absolute-path', self.abspath)
        if self.hash_:
            node.append(self.hash_.to_xml())
        return node

    def __repr__(self):
        return ('File(givenpath=%r, abspath=%r, hash_=%r)' %
                (self.givenpath, self.abspath, self.hash_))

    def __hash__(self):
        return hash(self.givenpath) ^ hash(self.abspath) ^ hash(self.hash_)

    def accept(self, visitor):
        visitor.visit_file(self)

class Hash(JsonMixin):
    attrs = [Attribute('alg', _string_type),
             Attribute('hexdigest',  _string_type)]

    def __init__(self, alg, hexdigest):
        assert isinstance(alg, _string_type)
        assert isinstance(hexdigest, _string_type)
        self.alg = alg
        self.hexdigest = hexdigest

    @classmethod
    def from_xml(cls, node):
        alg = node.get('alg')
        hexdigest = node.get('hexdigest')
        result = Hash(alg, hexdigest)
        return result

    def to_xml(self):
        node = ET.Element('hash')
        node.set('alg', self.alg)
        node.set('hexdigest', self.hexdigest)
        return node

    def __repr__(self):
        return ('Hash(alg=%r, hexdigest=%r)' %
                (self.alg, self.hexdigest))

    def __hash__(self):
        return hash(self.alg) ^ hash(self.hexdigest)

class Function(JsonMixin):
    attrs = [Attribute('name', _string_type)]

    def __init__(self, name):
        self.name = name

    @classmethod
    def from_xml(cls, node):
        name = node.get('name')
        result = Function(name)
        return result

    def to_xml(self):
        node = ET.Element('function')
        node.set('name', self.name)
        return node

    def __repr__(self):
        return 'Function(name=%r)' % self.name

    def __hash__(self):
        return hash(self.name)

    def accept(self, visitor):
        visitor.visit_function(self)

class Point(JsonMixin):
    attrs = [Attribute('line', int),
             Attribute('column', int)]

    def __init__(self, line, column):
        assert isinstance(line, int)
        assert isinstance(column, int)
        self.line = line
        self.column = column

    @classmethod
    def from_xml(cls, node):
        line = int(node.get('line'))
        column = int(node.get('column'))
        result = Point(line, column)
        return result

    def to_xml(self):
        node = ET.Element('point')
        node.set('line', str(self.line))
        node.set('column', str(self.column))
        return node

    def __repr__(self):
        return ('Point(line=%r, column=%r)' %
                (self.line, self.column))

    def __hash__(self):
        return hash(self.line) ^ hash(self.column)

    def accept(self, visitor):
        visitor.visit_point(self)

class Range(JsonMixin):
    attrs = [Attribute('start', 'Point'),
             Attribute('end', 'Point')]

    def __init__(self, start, end):
        assert isinstance(start, Point)
        assert isinstance(end, Point)
        self.start = start
        self.end = end

    @classmethod
    def from_xml(cls, node):
        children = list(node)
        start = Point.from_xml(children[0])
        end = Point.from_xml(children[1])
        result = Range(start, end)
        return result

    def to_xml(self):
        node = ET.Element('range')
        node.append(self.start.to_xml())
        node.append(self.end.to_xml())
        return node

    def __repr__(self):
        return ('Range(start=%r, end=%r)' %
                (self.start, self.end))

    def __hash__(self):
        return hash(self.start) ^ hash(self.end)

    def accept(self, visitor):
        visitor.visit_range(self)
        self.start.accept(visitor)
        self.end.accept(visitor)

class CustomFields(OrderedDict):
    @classmethod
    def from_xml(cls, node):
        kvs = []
        for child_node in node:
            if child_node.tag == 'str-field':
                value = child_node.text
                if value is None:
                    value = ''
            elif child_node.tag == 'int-field':
                value = int(child_node.text)
            else:
                raise ValueError('unrecognized element within'
                                 ' <custom-fields>: %s'
                                 % child_node.tag)
            key = child_node.get('name')
            if key is None:
                raise ValueError('missing "name" attribute within'
                                 ' <%s>' % child_node.tag)
            kvs.append( (key, value) )
        result = cls(kvs)
        return result

    def to_xml(self):
        node = ET.Element('custom-fields')
        for key, value in iteritems(self):
            if isinstance(value, _string_type):
                tag = 'str-field'
                text = value
            elif isinstance(value, integer_types):
                tag = 'int-field'
                text = str(value)
            else:
                raise TypeError('unhandled type within CustomFields instance')
            field_node = ET.Element(tag)
            field_node.set('name', key)
            field_node.text = text
            node.append(field_node)
        return node

    @classmethod
    def from_json(cls, jsonobj):
        if jsonobj is None:
            return None
        return cls(jsonobj)

    def to_json(self):
        return OrderedDict(self)

    def __hash__(self):
        # dicts are usually mutable, but it would be useful to hash
        # CustomFields instances (and assume they don't change from under
        # us)
        result = 0
        for key, value in self.iteritems():
            result ^= hash(key) ^ hash(value)
        return result

#
# Traversal of the report structure
#

class Visitor:
    def visit_analysis(self, analysis):
        pass

    def visit_warning(self, warning):
        pass

    def visit_failure(self, failure):
        pass

    def visit_info(self, info):
        pass

    def visit_metadata(self, metadata):
        pass

    def visit_generator(self, generator):
        pass

    def visit_sut(self, sut):
        pass

    def visit_stats(self, stats):
        pass

    def visit_message(self, message):
        pass

    def visit_notes(self, notes):
        pass

    def visit_state(self, state):
        pass

    def visit_location(self, location):
        pass

    def visit_file(self, file_):
        pass

    def visit_function(self, function):
        pass

    def visit_point(self, point):
        pass

    def visit_range(self, range_):
        pass

def main():
    for filename in sorted(glob.glob('examples/example-*.xml')):
        print('%s as gcc output:' % filename)
        with open(filename) as f:
            r = Analysis.from_xml(f)
            for w in r.results:
                if isinstance(w, Issue):
                    w.write_as_gcc_output(sys.stderr)
            sys.stderr.write(str(r.to_xml()))

if __name__ == '__main__':
    main()



import unittest

import mock

from firehose.parsers import frama_c
from firehose.model import Issue, Location, File, Point, \
    Function, Message, Sut, Result, Analysis

FAKE_SUT = Sut()


class TestParseWarning(unittest.TestCase):

    @staticmethod
    def create_mock_file(lines):
        mock_file = mock.MagicMock()
        mock_file.readlines.return_value = lines
        return mock_file

    def get_file_with_warning(self):
        return self.create_mock_file(["new.c:13:[kernel] warning: Body of function foo falls-through. Adding a return statement"])

    def get_file_without_warning(self):
        return self.create_mock_file(["[kernel] Parsing perft.c (with preprocessing)"])

    def test_parse_warning(self):
        mock_file = self.get_file_with_warning()
        ret = frama_c.parse_file(mock_file)
        self.assertIsInstance(ret, Analysis)
        self.assertIsInstance(ret.results[0], Result)
        self.assertEqual(len(ret.results), 1)

    def test_paths(self):
        mock_file = self.get_file_with_warning()
        analysis = frama_c.parse_file(mock_file)
        self.assertIsInstance(analysis, Analysis)
        self.assertEqual(len(analysis.results), 1)

        results = analysis.results[0]
        self.assertIsInstance(results.location, Location)
        self.assertIsInstance(results.location.file, File)
        self.assertIsInstance(results.location.point, Point)
        self.assertEqual(results.location.function, None)
        self.assertIsInstance(results.message, Message)
        self.assertEqual(results.location.line, 13)

        self.assertEqual(
            results.message.text, "Body of function foo falls-through. Adding a return statement")
        self.assertEqual(results.location.file.givenpath, "new.c")
        self.assertEqual(results.location.file.abspath, None)

    def test_empty(self):
        mock_file = self.create_mock_file([""])
        ret = frama_c.parse_file(mock_file)
        self.assertEqual(len(ret.results), 0)

    def test_new_line(self):
        mock_file = self.create_mock_file(["\n"])
        ret = frama_c.parse_file(mock_file)
        self.assertEqual(len(ret.results), 0)

    def test_no_warning(self):
        mock_file = self.get_file_without_warning()
        ret = frama_c.parse_file(mock_file)
        self.assertEqual(len(ret.results), 0)



#!/usr/bin/env python
#
#   Copyright 2017 David Carlos  <ddavidcarlos1392@gmail.com>
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
#   USA

import os
import unittest

from firehose.parsers.flawfinder import parse_file
from firehose.model import Analysis, Issue, Sut, Trace

class TestParseXml(unittest.TestCase):
    def parse_example(self, filename):
        try:
            path = os.path.join(os.path.dirname(__file__),
                    'example-output',
                    'flawfinder',
                    filename)
            with open(path) as infile:
                return parse_file(infile)
        except IOError:
            print("Example input not found")

    def test_flawfinder_report(self):
        a = self.parse_example('flawfinder-report-1')
        self.assertEqual(a.metadata.generator.name, 'flawfinder')
        self.assertEqual(a.metadata.generator.version, '1.31')
        self.assertEqual(a.metadata.sut, None)
        self.assertEqual(a.metadata.file_, None)
        self.assertEqual(a.metadata.stats, None)
        self.assertEqual(a.metadata.stats, None)

        self.assertEqual(len(a.results), 1804)

        w0 = a.results[0]
        self.assertEqual(w0.cwe, 78)
        expected_message = 'This causes a new program to execute ' \
                           'and is difficult to use safely (CWE-78). ' \
                           'try using a library call that implements ' \
                           'the same functionality if available.'
        self.assertEqual(w0.message.text, expected_message)
        self.assertEqual(w0.testid, 'shell')
        self.assertEqual(w0.location.file.givenpath , "./docs/examples/asiohiper.cpp")
        self.assertEqual(w0.location.point.line, 78)
        self.assertEqual(w0.location.point.column, 0)

        w3 = a.results[4]
        self.assertEqual(w3.location.file.givenpath , "./docs/examples/cookie_interface.c")
        self.assertEqual(w3.testid, 'format')
        some_w = a.results[1801]
        self.assertEqual(some_w.cwe, 126)
        self.assertEqual(some_w.testid, 'buffer')
        other_w = a.results[1802]
        self.assertEqual(other_w.cwe, None)
        self.assertEqual(other_w.location.file.givenpath,
                         "./tests/unit/unit1604.c")
        self.assertEqual(other_w.location.point.line, 49)
        self.assertEqual(other_w.location.point.column, 0)



#!/usr/bin/env python
import re
import sys

from firehose.model import Message, Point, \
    File, Location, Metadata, Generator, Issue, Analysis

# Parser for warnings emitted by frama-c
# Frama-c allows for multiple analysis, including the following:
#
# Aorai               verification of behavioral properties (experimental)
#                     (-aorai-h)
# Callgraph           automatically compute the callgraph of the program. Using
#                     Value might improve the precision of this plug-in (-cg-h)
# Dominators          Compute postdominators of statements (-dominators-h)
# From analysis       functional dependencies (-from-h)
# Inout               operational, imperative and all kinds of inputs/outputs
#                     (-inout-h)
# Metrics             syntactic metrics (-metrics-h)
# Obfuscator          objuscator for confidential code (-obfuscator-h)
# Occurrence          automatically computes where variables are used
#                     (-occurrence-h)
# Pdg                 Program Dependence Graph (-pdg-h)
# Postdominators      computing postdominators of statements
#                     (-postdominators-h)
# Report              Properties Status Report (experimental) (-report-h)
# Rtegen              generates annotations for runtime error checking and
#                     preconditions at call sites (-rte-h)
# Scope               data dependencies higher level functions (-scope-h)
# Security-slicing    security slicing (experimental, undocumented)
#                     (-security-slicing-h)
# Semantic constant folding  propagates constants semantically (-scf-h)
# Slicing             code slicer (-slicing-h)
# Sparecode           code cleaner (-sparecode-h)
# Users               function callees (-users-h)
# Value analysis      automatically computes variation domains for the
#                     variables of the program (-value-h)
# WP                  Weakest Precondition Calculus
#                     WP 0.9 for Magnesium-20151002 (-wp-h)
#
# The example log provided in tests/parsers/example-input/sparecode.log is
# from the sparecode analysis run on a trivial C chess engine program

SPARECODE_WARNING = ("^(?P<path>\S.*?):(?P<line>\d+):(?P<module>\S.*?) (?P<type>warning|note): (?P<message>.*?)$")
FRAMA_C_SPARECODE_PATTERN = re.compile(SPARECODE_WARNING)


def parse_file(data_file, sut=None, file_=None, stats=None):
    """
    for each line, the regex for SPARECODE_WARNING is matched

    :param data_file:   file object containing build log
    :type  data_file:   file

    :return:    Analysis instance
    """

    generator = Generator(name='frama-c')
    metadata = Metadata(generator, sut, file_, stats)
    analysis = Analysis(metadata, [])

    for line in data_file.readlines():
        match_warning = FRAMA_C_SPARECODE_PATTERN.match(line)

        if match_warning:
            issue = parse_warning(match_warning)
            analysis.results.append(issue)
    return analysis


def parse_warning(match_warning):
    """
    :param match_warning:  the matched object
    :type  match_warning:  SRE_Match
    :param sut:   metadata about the software-under-test
    :type  sut:   Sut

    :return:    Issue
    """
    message = Message(match_warning.group('message'))
    point = Point(int(match_warning.group('line')), 0)
    path = File(match_warning.group('path'), None)
    location = Location(file=path, function=None, point=point)
    return Issue(
        cwe=None, testid=None, location=location, message=message, notes=None,
        trace=None)


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("Provide a build log file path as the only argument")
    else:
        with open(sys.argv[1]) as data_file:
            analysis = parse_file(data_file)
            sys.stdout.write(str(analysis.to_xml()))
            sys.stdout.write('\n')



#   Copyright 2013, 2017 Red Hat, Inc.
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
#   USA

import os
import unittest

from firehose.parsers.clanganalyzer import parse_plist
from firehose.model import Analysis, Issue, Sut, Trace

class TestParsePlist(unittest.TestCase):
    def parse_example(self, filename):
        a = parse_plist(os.path.join(os.path.dirname(__file__),
                                     'example-output/clanganalyzer',
                                     filename),
                        file_=None,
                        stats=None)
        return a

    def test_example_001(self):
        a = self.parse_example('report-001.plist')
        self.assertEqual(a.metadata.generator.name, 'clang-analyzer')
        self.assertEqual(a.metadata.generator.version, None)

        self.assertEqual(len(a.results), 2)

        w0 = a.results[0]
        self.assertEqual(w0.cwe, None)
        self.assertEqual(w0.testid, 'Dead assignment')
        self.assertEqual(w0.message.text,
                         "Value stored to 'ret' is never read")
        self.assertEqual(w0.notes, None)
        self.assertEqual(w0.location.file.givenpath,
                         'python-ethtool/ethtool.c')
        self.assertEqual(w0.location.file.abspath, None)
        self.assertEqual(w0.location.function, None)
        self.assertEqual(w0.location.line, 130)
        self.assertEqual(w0.location.column, 2)
        self.assertNotEqual(w0.trace, None)
        self.assertEqual(len(w0.trace.states), 1)
        s0 = w0.trace.states[0]
        self.assertEqual(s0.location.file.givenpath,
                         'python-ethtool/ethtool.c')
        self.assertEqual(s0.location.file.abspath, None)
        self.assertEqual(s0.location.function.name, '')
        self.assertEqual(s0.location.line, 130)
        self.assertEqual(s0.location.column, 2)
        self.assertEqual(s0.notes.text,
                         "Value stored to 'ret' is never read")

    def test_example_002(self):
        a = self.parse_example('report-002.plist')
        self.assertEqual(a.metadata.generator.name, 'clang-analyzer')
        self.assertEqual(a.metadata.generator.version, None)

        self.assertEqual(len(a.results), 4)

        w0 = a.results[0]
        self.assertEqual(w0.testid, 'Dead assignment')
        self.assertEqual(w0.message.text,
                         "Value stored to 'error' is never read")
        self.assertEqual(w0.location.file.givenpath, 'search.c')
        self.assertEqual(w0.location.line, 454)
        self.assertEqual(w0.location.column, 3)

        w1 = a.results[1]
        self.assertEqual(w1.testid, 'Dead increment')
        self.assertEqual(w1.message.text,
                         "Value stored to 'pol_opt' is never read")
        self.assertEqual(w1.location.file.givenpath, 'search.c')
        self.assertEqual(w1.location.line, 824)
        self.assertEqual(w1.location.column, 2)

        w2 = a.results[2]
        self.assertEqual(w2.testid, 'Dereference of null pointer')
        self.assertEqual(w2.message.text,
                         "Access to field 'ob_refcnt' results in a dereference of a null pointer (loaded from variable 'dict')")
        self.assertEqual(w2.location.file.givenpath, 'search.c')
        self.assertEqual(w2.location.line, 215)
        self.assertEqual(w2.location.column, 2)
        trace2 = w2.trace
        self.assertIsInstance(trace2, Trace)

        self.assertEqual(len(trace2.states), 13)

        # s0 and s1 come from the first control edge in the input file:
        #         'path': [{'edges': [{'end': [{'col': 9,
        #                                       'file': 0,
        #                                       'line': 161},
        #                                      {'col': 9,
        #                                       'file': 0,
        #                                       'line': 161}],
        #                              'start': [{'col': 2,
        #                                         'file': 0,
        #                                         'line': 161},
        #                                        {'col': 2,
        #                                         'file': 0,
        #                                         'line': 161}]}],
        #                   'kind': 'control'},

        s0 = trace2.states[0]
        self.assertEqual(s0.location.point.line, 161)
        self.assertEqual(s0.location.point.column, 2)
        self.assertEqual(s0.location.range_, None)


        s1 = trace2.states[1]
        self.assertEqual(s1.location.point.line, 161)
        self.assertEqual(s1.location.point.column, 9)
        self.assertEqual(s1.location.range_, None)

        # s2 comes from the endpoint of the second control edge in the
        # input file (the startpoint == s1).
        #                  {'edges': [{'end': [{'col': 18,
        #                                       'file': 0,
        #                                       'line': 165},
        #                                      {'col': 21,
        #                                       'file': 0,
        #                                       'line': 165}],
        #                              'start': [{'col': 9,
        #                                         'file': 0,
        #                                         'line': 161},
        #                                        {'col': 9,
        #                                         'file': 0,
        #                                         'line': 161}]}],
        #                   'kind': 'control'},
        #
        # It is a range rather than a point:
        s2 = trace2.states[2]
        self.assertEqual(s2.location.point, None)
        self.assertEqual(s2.location.range_.start.line, 165)
        self.assertEqual(s2.location.range_.start.column, 18)
        self.assertEqual(s2.location.range_.end.line, 165)
        self.assertEqual(s2.location.range_.end.column, 21)

        # s3 comes from the next entry in the input file, which is the first
        # "event" in the trace:
        #                  {'extended_message': "Variable 'dict' initialized to a null pointer value",
        #                   'kind': 'event',
        #                   'location': {'col': 18,
        #                                'file': 0,
        #                                'line': 165},
        #                   'message': "Variable 'dict' initialized to a null pointer value",
        #                   'ranges': [[{'col': 18,
        #                                'file': 0,
        #                                'line': 165},
        #                               {'col': 21,
        #                                'file': 0,
        #                                'line': 165}]]},
        s3 = trace2.states[3]
        # The importer uses the 'location' point for the event, and hence
        # is treated as a different location.  However, as an event, it is
        # always given its own state in the imported data:
        self.assertEqual(s3.location.point.line, 165)
        self.assertEqual(s3.location.point.column, 18)
        self.assertEqual(s3.notes.text,
                         "Variable 'dict' initialized to a null pointer value")
        self.assertEqual(s3.location.range_, None)

        # The next entry in the input file is another kind == 'control':
        #                  {'edges': [{'end': [{'col': 2,
        #                                       'file': 0,
        #                                       'line': 171},
        #                                      {'col': 2,
        #                                       'file': 0,
        #                                       'line': 171}],
        #                              'start': [{'col': 18,
        #                                         'file': 0,
        #                                         'line': 165},
        #                                        {'col': 21,
        #                                         'file': 0,
        #                                         'line': 165}]}],
        #                   'kind': 'control'},
        # The "start" is range-based, whereas the previous event was
        # handled as a point, so the importer will treat it as two
        # different locations (s3 and s4):
        s4 = trace2.states[4]
        self.assertEqual(s4.location.point, None)
        self.assertEqual(s4.location.range_.start.line, 165)
        self.assertEqual(s4.location.range_.start.column, 18)
        self.assertEqual(s4.location.range_.end.line, 165)
        self.assertEqual(s4.location.range_.end.column, 21)

        s5 = trace2.states[5]
        self.assertEqual(s5.location.point.line, 171)
        self.assertEqual(s5.location.point.column, 2)
        self.assertEqual(s5.location.range_, None)

        # etc

    def test_example_003(self):
        a = self.parse_example('report-003.plist')
        self.assertEqual(a.metadata.generator.name, 'clang-analyzer')

        # This example has a version
        self.assertEqual(a.metadata.generator.version,
                         'clang version 3.4.2 (tags/RELEASE_34/dot2-final)')

        self.assertEqual(len(a.results), 1)

        w0 = a.results[0]
        self.assertEqual(w0.testid, 'Garbage return value')
        self.assertEqual(w0.message.text,
                         "Undefined or garbage value returned to caller")
        self.assertEqual(w0.location.file.givenpath,
                         '../../src/test-sources/out-of-bounds.c')
        self.assertEqual(w0.location.line, 5)
        self.assertEqual(w0.location.column, 3)

        # Verify that we capture various clang-specific per-issue metadata
        self.assertEqual(w0.customfields['category'], 'Logic error')
        self.assertEqual(w0.customfields['issue_context'], 'out_of_bounds')
        self.assertEqual(w0.customfields['issue_context_kind'], 'function')

if __name__ == '__main__':
    unittest.main()



#   Copyright 2017 David Malcolm <dmalcolm@redhat.com>
#   Copyright 2017 Red Hat, Inc.
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
#   USA

import os
import unittest

from firehose.model import Issue

from firehose.parsers.splint import parse_splint_csv, parse_splint_stderr

class TestParser(unittest.TestCase):
    def locate_filename(self, filename):
        return os.path.join(os.path.dirname(__file__),
                            'example-output',
                            'splint',
                            filename)

    def parse_example_csv(self, filename):
        return parse_splint_csv(self.locate_filename(filename))

    def test_unconditional_file_leak(self):
        a = self.parse_example_csv('unconditional-file-leak.csv')
        self.assertEqual(a.metadata.generator.name, 'splint')
        self.assertEqual(a.metadata.generator.version, None) # FIXME
        self.assertEqual(a.metadata.sut, None)
        self.assertEqual(a.metadata.file_, None)
        self.assertEqual(a.metadata.stats, None)
        self.assertEqual(len(a.results), 8)
        r0 = a.results[0]
        self.assertIsInstance(r0, Issue)
        self.assertEqual(r0.cwe, None)
        self.assertEqual(r0.testid, 'internalglobs')
        self.assertEqual(r0.message.text,
                         'Called procedure fopen may access file system'
                         ' state, but globals list does not include globals'
                         ' fileSystem')
        self.assertEqual(r0.notes.text,
                         'A called function uses internal state, but the'
                         ' globals list for the function being checked does'
                         ' not include internalState')
        self.assertEqual(r0.location.file.givenpath,
                         'examples/unconditional-file-leak.c')
        self.assertEqual(r0.location.file.abspath, None)
        self.assertEqual(r0.location.function, None)
        self.assertEqual(r0.trace, None)
        self.assertEqual(r0.severity, '1')

        # Verify that rows with unescaped quotes are correctly worked around
        # In this example, this affects warnings #4 and #6 (these are 1-based,
        # so index 3 and 5 in the list).
        r3 = a.results[3]
        self.assertEqual(r3.location.line, 9)
        self.assertEqual(r3.location.column, 5)
        # Quote-handling is not quite perfect, we've lost the open quote of
        # the format string, and there an erroneous trailing quote in this
        # message.
        self.assertEqual(r3.message.text,
                         'Undocumented modification of file system state'
                         ' possible from call to fprintf:'
                         ' fprintf(f, %i: %i", i, i * i)"')

        r5 = a.results[5]
        # Similar issues here:
        self.assertEqual(r5.message.text,
                         'Body of for statement is not a block:'
                         ' fprintf(f, %i: %i", i, i * i);"')

    def parse_example_stderr(self, filename):
        with open(self.locate_filename(filename)) as f:
            stderr = f.read()
        return parse_splint_stderr(stderr)

    def test_parse_stderr(self):
        v = self.parse_example_stderr('unconditional-file-leak.stderr')
        self.assertEqual(v, '3.1.2')



#   Copyright 2013 Shuxiong Ye <yeshuxiong@gmail.com> 
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301 USA

import unittest
import os

from firehose.parsers import findbugs# import parse_warning, parse_file
from firehose.model import Analysis, Issue, Location, File, Point, \
    Function, Message, Sut, Metadata, Generator

FAKE_SUT = Sut()

def readRawData(filePath):
    f = open(filePath, "r")
    ret = f.readlines()
    f.close()
    return ret

class TestParseFile(unittest.TestCase):
    def run_testdata(self, xmlFileName, answer, isDebug=False):
        path = os.path.join(os.path.dirname(__file__), "example-output",
                "findbugs", xmlFileName)
        analysis = findbugs.parse_file(path)
        self.assertIsInstance(analysis, Analysis)
        self.assertNotEqual(analysis.results, None)
        if isDebug:
            print(len(analysis.results), len(answer))
        self.assertEqual(len(analysis.results), len(answer))
        for i in range(0, len(analysis.results)):
            ret = analysis.results[i] 
            self.assertIsInstance(ret, Issue)
            self.assertIsInstance(ret.location, Location)
            self.assertEqual(ret.location.line, answer[i][0])
            if answer[i][1]:
                self.assertIsInstance(ret.location.function, Function)
                self.assertEqual(ret.location.function.name, answer[i][1])
            else:
                self.assertEqual(ret.location.function, None)
            self.assertIsInstance(ret.location.file, File)
            self.assertEqual(ret.location.file.givenpath, answer[i][2])

    def test_testdata_jformatstring(self):
        answer = [
                [66, "edu.umd.cs.findbugs.formatStringChecker."+
                "FormatterCompileTimeTest.testBug1874856FalsePositive()",
                "edu/umd/cs/findbugs/formatStringChecker/"+
                "FormatterCompileTimeTest.java"],
                [89, "edu.umd.cs.findbugs.formatStringChecker."+
                "FormatterCompileTimeTest.testBug1874856TruePositive()",
                "edu/umd/cs/findbugs/formatStringChecker/"+
                "FormatterCompileTimeTest.java"],
                [19, "edu.umd.cs.findbugs.formatStringChecker."+
                "FormatterRuntimeTest.testShouldWork()",
                "edu/umd/cs/findbugs/formatStringChecker/"+
                "FormatterRuntimeTest.java"],
                [52, "edu.umd.cs.findbugs.formatStringChecker."+
                "FormatterRuntimeTest.testFormatDateWithY()",
                "edu/umd/cs/findbugs/formatStringChecker/"+
                "FormatterRuntimeTest.java"],
                [36, "edu.umd.cs.findbugs.formatStringChecker."+
                "FormatterRuntimeTest.testNotEnoughParameters()",
                "edu/umd/cs/findbugs/formatStringChecker/"+
                "FormatterRuntimeTest.java"],
                [28, "edu.umd.cs.findbugs.formatStringChecker."+
                "FormatterRuntimeTest.testStringWhereIntegerExpected()",
                "edu/umd/cs/findbugs/formatStringChecker/"+
                "FormatterRuntimeTest.java"],
                [88, "edu.umd.cs.findbugs.formatStringChecker."+
                "FormatterRuntimeTest.testDateMismatch()",
                "edu/umd/cs/findbugs/formatStringChecker/"+
                "FormatterRuntimeTest.java"],
                [28, "edu.umd.cs.findbugs.formatStringChecker."+
                "FormatterRuntimeTest.testStringWhereIntegerExpected()",
                "edu/umd/cs/findbugs/formatStringChecker/"+
                "FormatterRuntimeTest.java"],
                [43, "edu.umd.cs.findbugs.formatStringChecker."+
                "FormatterRuntimeTest.testPassingAnArray()",
                "edu/umd/cs/findbugs/formatStringChecker/"+
                "FormatterRuntimeTest.java"],
                [47, "edu.umd.cs.findbugs.formatStringChecker."+
                "FormatterRuntimeTest.testPassingAnIntToABoolean()",
                "edu/umd/cs/findbugs/formatStringChecker/"+
                "FormatterRuntimeTest.java"],
                [52, "edu.umd.cs.findbugs.formatStringChecker."+
                "FormatterRuntimeTest.testFormatDateWithY()",
                "edu/umd/cs/findbugs/formatStringChecker/"+
                "FormatterRuntimeTest.java"],
                [36, "edu.umd.cs.findbugs.formatStringChecker."+
                "FormatterRuntimeTest.testNotEnoughParameters()",
                "edu/umd/cs/findbugs/formatStringChecker/"+
                "FormatterRuntimeTest.java"],
                [78, "edu.umd.cs.findbugs.formatStringChecker."+
                "FormatterRuntimeTest.testBug1874856TruePositive()",
                "edu/umd/cs/findbugs/formatStringChecker/"+
                "FormatterRuntimeTest.java"],
                [88, "edu.umd.cs.findbugs.formatStringChecker."+
                "FormatterRuntimeTest.testDateMismatch()",
                "edu/umd/cs/findbugs/formatStringChecker/"+
                "FormatterRuntimeTest.java"],
                [21, "edu.umd.cs.findbugs.formatStringChecker"+
                ".FormatterRuntimeTest.testShouldWork()",
                "edu/umd/cs/findbugs/formatStringChecker/"+
                "FormatterRuntimeTest.java"],
                [22, "edu.umd.cs.findbugs.formatStringChecker."+
                "FormatterRuntimeTest.testShouldWork()",
                "edu/umd/cs/findbugs/formatStringChecker/"+
                "FormatterRuntimeTest.java"]
                ]
        self.run_testdata("findbugs_jformatstring.xml", answer)

    def test_testdata_example(self):
        answer = [
                [8, "O0.<static initializer for O0>()", "O0.java"],
                [13, "new O0()", "O0.java"],
                [18, "O0.bugs()", "O0.java"],
                [23, "O0.main(String[])", "O0.java"],
                [9, "l1.O1.<static initializer for O1>()", "l1/O1.java"],
                [14, "new l1.O1()", "l1/O1.java"],
                [19, "l1.O1.bugs()", "l1/O1.java"],
                [24, "l1.O1.main(String[])", "l1/O1.java"],
                [9, "l1.l2.O2.<static initializer for O2>()",
                    "l1/l2/O2.java"],
                [14, "new l1.l2.O2()", "l1/l2/O2.java"],
                [19, "l1.l2.O2.bugs()", "l1/l2/O2.java"],
                [24, "l1.l2.O2.main(String[])", "l1/l2/O2.java"],
                ]
        self.run_testdata("findbugs_Example.xml", answer)

    def test_testdata_no_method(self):
        answer = [
                [77, None, "com/clearspring/ircbot/trac/TracRpc.java"],
                ]
        self.run_testdata("findbugs_no_method.xml", answer)



#   Copyright 2013, 2017 David Malcolm <dmalcolm@redhat.com>
#   Copyright 2013, 2017 Red Hat, Inc.
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
#   USA

# Parser for the .plist files emitted by the clang-static-analyzer,
# when "-plist" is passed as an option to "scan-build" or "clang"

# Originally developed against output from clang-3.0-14.fc17;
# updated against output from clang-3.4-12.fc20.x86_64

import glob
import os
import plistlib
from pprint import pprint
import sys

from firehose.model import Message, Function, Point, Range, \
    File, Location, Generator, Metadata, Analysis, Issue, Sut, Trace, \
    State, Notes, CustomFields

def parse_scandir(resultdir, analyzerversion=None, sut=None):
    """
    Given a path to a directory of scan-build output, parse it and
    yield Analysis instances
    """
    for filename in glob.glob(os.path.join(resultdir, 'report-*.plist')):
        yield parse_plist(filename, analyzerversion, sut)

def parse_plist(pathOrFile, analyzerversion=None, sut=None, file_=None, stats=None):
    """
    Given a .plist file emitted by clang-static-analyzer (e.g. via
    scan-build), parse it and return an Analysis instance
    """
    plist = plistlib.readPlist(pathOrFile)
    # We now have the .plist file as a hierarchy of dicts, lists, etc

    # Handy debug dump:
    if 0:
        pprint(plist)

    # A list of filenames, apparently referenced by index within
    # diagnostics:
    files = plist['files']

    generator = Generator(name='clang-analyzer',
                          version=analyzerversion)
    metadata = Metadata(generator, sut, file_, stats)
    analysis = Analysis(metadata, [])

    if 'clang_version' in plist:
        generator.version = plist['clang_version']

    for diagnostic in plist['diagnostics']:
        if 0:
            pprint(diagnostic)

        cwe = None

        customfields = CustomFields()
        for key in ['category', 'issue_context', 'issue_context_kind']:
            if key in diagnostic:
                customfields[key] = diagnostic[key]

        message = Message(text=diagnostic['description'])

        loc = diagnostic['location']
        location = Location(file=File(givenpath=files[loc['file']],
                                      abspath=None),

                            # FIXME: doesn't tell us function name
                            # TODO: can we patch this upstream?
                            function=None,

                            point=Point(int(loc['line']),
                                        int(loc['col'])))

        notes = None

        trace = make_trace(files, diagnostic['path'])

        issue = Issue(cwe,
                      # Use the 'type' field for the testid:
                      diagnostic['type'],
                      location, message, notes, trace,
                      customfields=customfields)

        analysis.results.append(issue)

    return analysis

def make_point_from_plist_point(loc):
    # point:
    #   e.g. {'col': 2, 'file': 0, 'line': 130}
    return Point(int(loc['line']),
                 int(loc['col']))

def make_location_from_point(files, loc):
    # loc:
    #   e.g. {'col': 2, 'file': 0, 'line': 130}
    location = Location(file=File(givenpath=files[loc['file']],
                                  abspath=None),

                        # FIXME: doesn't tell us function name
                        # TODO: can we patch this upstream?
                        function=Function(''),

                        point=make_point_from_plist_point(loc))
    return location

def make_location_from_range(files, range_):
    # range_:
    #    e.g.:
    #     [{'col': 18, 'file': 0, 'line': 165},
    #      {'col': 21, 'file': 0, 'line': 165}]
    assert len(range_) == 2
    start = range_[0]
    end = range_[1]
    assert start['file'] == end['file']

    if start == end:
        point = make_point_from_plist_point(start)
        range_ = None
    else:
        point = None
        range_ = Range(start=make_point_from_plist_point(start),
                       end=make_point_from_plist_point(end))

    location = Location(file=File(givenpath=files[start['file']],
                                  abspath=None),

                        # FIXME: doesn't tell us function name
                        # TODO: can we patch this upstream?
                        function=Function(''),

                        point=point,
                        range_=range_)

    return location

def make_trace(files, path):
    """
    Construct a Trace instance from the .plist's 'path' list
    """
    trace = Trace([])
    lastlocation = None
    for node in path:
        if 0:
            pprint(node)

        kind = node['kind']

        if kind == 'event':
            # e.g.:
            #  {'extended_message': "Value stored to 'ret' is never read",
            #   'kind': 'event',
            #   'location': {'col': 2, 'file': 0, 'line': 130},
            #   'message': "Value stored to 'ret' is never read",
            #   'ranges': [[{'col': 8, 'file': 0, 'line': 130},
            #               {'col': 29, 'file': 0, 'line': 130}]]}

            # TODO: we're not yet handling the following:
            #   node['extended_message']
            #   node['ranges']

            loc = node['location']
            location = make_location_from_point(files, loc)

            notes = Notes(node['message'])
            trace.add_state(State(location, notes))

            lastlocation = location

        elif kind == 'control':
            # e.g.:
            #  {'edges': [{'end': [{'col': 9, 'file': 0, 'line': 161},
            #                      {'col': 9, 'file': 0, 'line': 161}],
            #              'start': [{'col': 2, 'file': 0, 'line': 161},
            #                        {'col': 2, 'file': 0, 'line': 161}]}],
            #   'kind': 'control'}
            edges = node['edges']
            for edge in edges:
                edge_start = edge['start']
                edge_end = edge['end']

                startloc = make_location_from_range(files, edge_start)
                endloc = make_location_from_range(files, edge_end)

                if startloc != lastlocation:
                    trace.add_state(State(startloc, None))
                trace.add_state(State(endloc, None))
                lastlocation = endloc
        else:
            raise ValueError('unknown kind: %r' % kind)
    return trace

if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("provide either the path to scan-build results directory, or to that of a .plist file as the only argument")
    else:
        path = sys.argv[1]
        if path.endswith('.plist'):
            analysis = parse_plist(path)
            sys.stdout.write(str(analysis.to_xml()))
            sys.stdout.write('\n')
        else:
            for analysis in parse_scandir(path):
                sys.stdout.write(str(analysis.to_xml()))
                sys.stdout.write('\n')



#!/usr/bin/env python

#   Copyright 2013 Red Hat, Inc.
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
#   USA

import re
import sys

from firehose.model import Message, Function, Point, \
    File, Location, Metadata, Generator, Issue, Analysis

# Parser for warnings emitted by GCC
# The code that generates these warnings can be seen within gcc's own
# sources within:
#   gcc/diagnostic.c
#   gcc/langhooks.c: lhd_print_error_function
# (as of gcc-4.7.2)
# See e.g.:
#   http://gcc.gnu.org/viewcvs/trunk/gcc/diagnostic.c?revision=195098&view=markup
#   http://gcc.gnu.org/viewcvs/trunk/gcc/langhooks.c?revision=195098&view=markup
# This parser is only intended to be run with the C locale

# column is optional
# switch is optional
GCC_PATTERN = re.compile("^(?P<path>\S.*?):(?P<line>\d+):(?P<column>\d*):? (?P<type>warning|note): (?P<message>.*?)(?P<switch> \[\-W.+\])?$")

SWITCH_SUB_PATTERN = re.compile("^ \[\-W(?P<name>.*)\]$")

# single quotes may not match locales that are not C
FUNCTION_PATTERN = re.compile(".*: In (?:member )?function '(?P<func>.*)':")

# match when gcc issues a warning for a location it thinks is in global scope
GLOBAL_PATTERN = re.compile(".*: At global scope:$")

# When gcc issues a warning at spot it thinks is in global scope, use this
# as the function name
GLOBAL_FUNC_NAME = '::'


def parse_file(data_file, gccversion=None, sut=None, file_=None, stats=None):
    """
    looks for groups of lines that start with a line identifying a function
    name, followed by one or more lines with a warning or note

    :param data_file:   file object containing build log
    :type  data_file:   file
    :param gccversion:   version of GCC that generated this report
    :type  gccversion:   str

    :return:    Analysis instance
    """
    # has a value only when in a block of lines where the first line identifies
    # a function and is followed by 0 or more warning lines

    generator = Generator(name='gcc',
                          version=gccversion)
    metadata = Metadata(generator, sut, file_, stats)
    analysis = Analysis(metadata, [])

    current_func_name = None
    for line in data_file.readlines():
        match_func = FUNCTION_PATTERN.match(line)
        match_global = GLOBAL_PATTERN.match(line)
        # if we found a line that describes a function name
        if match_func:
            current_func_name = match_func.group('func')
        elif match_global:
            current_func_name = GLOBAL_FUNC_NAME

        # if we think the next line might describe a warning
        elif current_func_name is not None:
            issue = parse_warning(line, current_func_name)
            if issue:
                analysis.results.append(issue)
            else:
                # reset this when we run out of warnings associated with it
                current_func_name = None
    return analysis
                
            
def parse_warning(line, func_name):
    """
    :param line:        current line read from file
    :type  line:        basestring
    :param func_name:   name of the current function
    :type  func_name:   basestring
    :param gccversion:   version of GCC that generated this report
    :type  gccversion:   str
    :param sut:   metadata about the software-under-test
    :type  sut:   Sut

    :return:    Issue if match, else None
    """
    match = GCC_PATTERN.match(line)
    if match:
        message = Message(match.group('message'))
        func = Function(func_name)
        try:
            column = int(match.group('column'))
        except ValueError:
            if match.group('column') == '':
                column = 0
            else:
                raise
        except TypeError:
            column = None
        switch_match = SWITCH_SUB_PATTERN.match(match.group('switch') or '')
        if switch_match:
            switch = switch_match.group('name')
        else:
            switch = None

        point = Point(int(match.group('line')), column)
        path = File(match.group('path'), None)
        location = Location(path, func, point)

        return Issue(None, switch, location, message, None, None)


if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("provide a build log file path as the only argument")
    else:
        with open(sys.argv[1]) as data_file:
            analysis = parse_file(data_file)
            sys.stdout.write(str(analysis.to_xml()))
            sys.stdout.write('\n')



#   Copyright 2013 David Malcolm <dmalcolm@redhat.com>
#   Copyright 2013 Red Hat, Inc.
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
#   USA

import glob
import os
import subprocess
import tempfile
import unittest

from six import u, StringIO, BytesIO

from firehose.model import Analysis, Issue, Metadata, Generator, SourceRpm, \
    Location, File, Function, Point, Message, Notes, Trace, State, Stats, \
    Failure, Range, DebianSource, DebianBinary, CustomFields, Info

class AnalysisTests(unittest.TestCase):
    def make_simple_analysis(self):
        """
        Construct a minimal Analysis instance
        """
        a = Analysis(metadata=Metadata(generator=Generator(name='cpychecker'),
                                       sut=None,
                                       file_=None,
                                       stats=None),
                     results=[Issue(cwe=None,
                                    testid=None,
                                    location=Location(file=File('foo.c', None),
                                                      function=None,
                                                      point=Point(10, 15)),
                                    message=Message(text='something bad involving pointers'),
                                    notes=None,
                                    trace=None)])
        return a, a.results[0]

    def make_complex_analysis(self):
        """
        Construct a Analysis instance that uses all features
        """
        a = Analysis(metadata=Metadata(generator=Generator(name='cpychecker',
                                                           version='0.11'),
                                       sut=SourceRpm(name='python-ethtool',
                                                     version='0.7',
                                                     release='4.fc19',
                                                     buildarch='x86_64'),
                                       file_=File(givenpath='foo.c',
                                                  abspath='/home/david/coding/foo.c'),
                                       stats=Stats(wallclocktime=0.4)),
                     results=[Issue(cwe=681,
                                    testid='refcount-too-high',
                                    location=Location(file=File(givenpath='foo.c',
                                                                abspath='/home/david/coding/foo.c'),
                                                      function=Function('bar'),
                                                      point=Point(10, 15)),
                                    message=Message(text='something bad involving pointers'),
                                    notes=Notes('here is some explanatory text'),
                                    trace=Trace([State(location=Location(file=File('foo.c', None),
                                                                         function=Function('bar'),
                                                                         point=Point(7, 12)),
                                                       notes=Notes('first we do this')),
                                                 State(location=Location(file=File('foo.c', None),
                                                                         function=Function('bar'),
                                                                         point=Point(8, 10)),
                                                       notes=Notes('then we do that')),
                                                 State(location=Location(file=File('foo.c', None),
                                                                         function=Function('bar'),
                                                                         range_=Range(Point(10, 15),
                                                                                      Point(10, 25))),
                                                       notes=Notes('then it crashes here'))
                                                 ]),
                                    severity='really bad',
                                    customfields=CustomFields(foo='bar')),
                              ],
                     customfields=CustomFields(gccinvocation='gcc -I/usr/include/python2.7 -c foo.c'),
                     )
        return a, a.results[0]

    def make_failed_analysis(self):
        a = Analysis(metadata=Metadata(generator=Generator(name='yet-another-checker'),
                                       sut=None,
                                       file_=None,
                                       stats=None),
                     results=[Failure(failureid='out-of-memory',
                                      location=Location(file=File('foo.c', None),
                                                        function=Function('something_complicated'),
                                                        point=Point(10, 15)),
                                      message=Message('out of memory'),
                                      customfields=CustomFields(stdout='sample stdout',
                                                                stderr='sample stderr',
                                                                returncode=-9)) # (killed)
                              ])
        return a, a.results[0]

    def make_info(self):
        a = Analysis(metadata=Metadata(generator=Generator(name='an-invented-checker'),
                                       sut=None,
                                       file_=None,
                                       stats=None),
                     results=[Info(infoid='gimple-stats',
                                   location=Location(file=File('bar.c', None),
                                                     function=Function('sample_function'),
                                                     point=Point(10, 15)),
                                   message=Message('sample message'),
                                   customfields=CustomFields(num_stmts=57,
                                                             num_basic_blocks=10))
                              ])
        return a, a.results[0]

    def test_creating_simple_analysis(self):
        a, w = self.make_simple_analysis()
        self.assertEqual(a.metadata.generator.name, 'cpychecker')
        self.assertEqual(a.metadata.generator.version, None)
        self.assertEqual(a.metadata.sut, None)
        self.assertEqual(a.metadata.file_, None)
        self.assertEqual(a.metadata.stats, None)
        self.assertEqual(w.cwe, None)
        self.assertEqual(w.testid, None)
        self.assertEqual(w.location.file.givenpath, 'foo.c')
        self.assertEqual(w.location.file.abspath, None)
        self.assertEqual(w.location.function, None)
        self.assertEqual(w.location.line, 10)
        self.assertEqual(w.location.column, 15)
        self.assertEqual(w.message.text, 'something bad involving pointers')
        self.assertEqual(w.notes, None)
        self.assertEqual(w.trace, None)

    def test_creating_complex_analysis(self):
        a, w = self.make_complex_analysis()
        self.assertEqual(a.metadata.generator.name, 'cpychecker')
        self.assertEqual(a.metadata.generator.version, '0.11')
        self.assertIsInstance(a.metadata.sut, SourceRpm)
        self.assertEqual(a.metadata.sut.name, 'python-ethtool')
        self.assertEqual(a.metadata.sut.version, '0.7')
        self.assertEqual(a.metadata.sut.release, '4.fc19')
        self.assertEqual(a.metadata.sut.buildarch, 'x86_64')
        self.assertEqual(a.metadata.file_.givenpath, 'foo.c')
        self.assertEqual(a.metadata.file_.abspath, '/home/david/coding/foo.c')
        self.assertEqual(a.metadata.stats.wallclocktime, 0.4)
        self.assertEqual(w.cwe, 681)
        self.assertEqual(w.testid, 'refcount-too-high')
        self.assertEqual(w.location.file.givenpath, 'foo.c')
        self.assertEqual(w.location.file.abspath, '/home/david/coding/foo.c')
        self.assertEqual(w.location.function.name, 'bar')
        self.assertEqual(w.location.line, 10)
        self.assertEqual(w.location.column, 15)
        self.assertEqual(w.message.text, 'something bad involving pointers')
        self.assertEqual(w.notes.text, 'here is some explanatory text')
        self.assertEqual(w.severity, 'really bad')

        self.assertIsInstance(w.trace, Trace)
        self.assertEqual(len(w.trace.states), 3)
        s0 = w.trace.states[0]
        self.assertIsInstance(s0, State)
        self.assertEqual(s0.location.file.givenpath, 'foo.c')
        self.assertEqual(s0.location.function.name, 'bar')
        self.assertEqual(s0.location.line, 7)
        self.assertEqual(s0.location.column, 12)
        self.assertEqual(s0.notes.text, 'first we do this')

        # Verify the Range type within the final state in the trace:
        s2 = w.trace.states[2]
        self.assertIsInstance(s2, State)
        self.assertEqual(s2.location.line, 10)
        self.assertEqual(s2.location.column, 15)

    def test_making_failed_analysis(self):
        a, f = self.make_failed_analysis()

        self.assertIsInstance(f, Failure)
        self.assertEqual(f.failureid, 'out-of-memory')
        self.assertEqual(f.location.file.givenpath, 'foo.c')
        self.assertEqual(f.location.function.name, 'something_complicated')
        self.assertEqual(f.location.line, 10)
        self.assertEqual(f.location.column, 15)
        self.assertEqual(f.message.text, 'out of memory')
        self.assertEqual(f.customfields['stdout'], 'sample stdout')
        self.assertEqual(f.customfields['stderr'], 'sample stderr')
        self.assertEqual(f.customfields['returncode'], -9)

    def test_making_info(self):
        a, info = self.make_info()

        self.assertIsInstance(info, Info)
        self.assertEqual(info.infoid, 'gimple-stats')
        self.assertEqual(info.location.file.givenpath, 'bar.c')
        self.assertEqual(info.location.function.name, 'sample_function')
        self.assertEqual(info.location.line, 10)
        self.assertEqual(info.location.column, 15)
        self.assertEqual(info.message.text, 'sample message')
        self.assertEqual(info.customfields['num_stmts'], 57)
        self.assertEqual(info.customfields['num_basic_blocks'], 10)

    def test_from_xml(self):
        num_analyses = 0
        for filename in sorted(glob.glob('examples/example-*.xml')):
            with open(filename) as f:
                r = Analysis.from_xml(f)
                num_analyses += 1
        # Ensure that all of the reports were indeed parsed:
        self.assertEqual(num_analyses, 10)

    def test_example_2(self):
        # Verify that the parser works:
        with open('examples/example-2.xml') as f:
            a = Analysis.from_xml(f)
            self.assertEqual(a.metadata.generator.name, 'cpychecker')
            self.assertEqual(a.metadata.generator.version, '0.11')
            self.assertIsInstance(a.metadata.sut, SourceRpm)
            self.assertEqual(a.metadata.sut.name, 'python-ethtool')
            self.assertEqual(a.metadata.sut.version, '0.7')
            self.assertEqual(a.metadata.sut.release, '4.fc19')
            self.assertEqual(a.metadata.sut.buildarch, 'x86_64')

            self.assertEqual(len(a.results), 1)
            w = a.results[0]
            self.assertIsInstance(w, Issue)
            self.assertEqual(w.cwe, 401)
            self.assertEqual(w.testid, 'refcount-too-high')
            self.assertEqual(w.location.file.givenpath, 'examples/python-src-example.c')
            self.assertEqual(w.location.file.abspath, None)
            self.assertEqual(w.location.file.hash_.alg, 'sha1')
            self.assertEqual(w.location.file.hash_.hexdigest,
                             '6ba29daa94d64b48071e299a79f2a00dcd99eeb1')
            self.assertEqual(w.location.function.name, 'make_a_list_of_random_ints_badly')
            self.assertEqual(w.location.line, 40)
            self.assertEqual(w.location.column, 4)
            self.assertEqual(w.message.text, "ob_refcnt of '*item' is 1 too high")
            self.assertMultiLineEqual(w.notes.text,
                ("was expecting final item->ob_refcnt to be N + 1 (for some unknown N)\n"
                 "due to object being referenced by: PyListObject.ob_item[0]\n"
                 "but final item->ob_refcnt is N + 2"))

            self.assertIsInstance(w.trace, Trace)
            self.assertEqual(len(w.trace.states), 3)
            s0 = w.trace.states[0]
            self.assertIsInstance(s0, State)
            self.assertEqual(s0.location.file.givenpath, 'examples/python-src-example.c')
            self.assertEqual(s0.location.function.name, 'make_a_list_of_random_ints_badly')
            self.assertEqual(s0.location.line, 36)
            self.assertEqual(s0.location.column, 14)
            self.assertEqual(s0.notes.text,
                'PyLongObject allocated at:         item = PyLong_FromLong(random());')

    def test_example_3(self):
        # Verify that the parser works:
        with open('examples/example-3.xml') as f:
            a = Analysis.from_xml(f)
            self.assertEqual(a.metadata.generator.name, 'cpychecker')
            self.assertEqual(a.metadata.generator.version, '0.11')
            self.assertIsInstance(a.metadata.sut, SourceRpm)
            self.assertEqual(a.metadata.sut.name, 'python-ethtool')
            self.assertEqual(a.metadata.sut.version, '0.7')
            self.assertEqual(a.metadata.sut.release, '4.fc19')
            self.assertEqual(a.metadata.sut.buildarch, 'x86_64')

            self.assertEqual(len(a.results), 1)
            w = a.results[0]
            self.assertIsInstance(w, Failure)
            self.assertEqual(w.failureid, 'bad-exit-code')
            self.assertEqual(w.customfields['returncode'], -11)

    def test_example_4(self):
        with open('examples/example-4.xml') as f:
            a = Analysis.from_xml(f)
            self.assertEqual(a.metadata.generator.name, 'cpychecker')
            self.assertEqual(a.metadata.generator.version, '0.11')
            self.assertIsInstance(a.metadata.sut, SourceRpm)
            self.assertEqual(a.metadata.sut.name, 'python-ethtool')
            self.assertEqual(a.metadata.sut.version, '0.7')
            self.assertEqual(a.metadata.sut.release, '4.fc19')
            self.assertEqual(a.metadata.sut.buildarch, 'x86_64')

            self.assertEqual(len(a.results), 1)
            w = a.results[0]
            self.assertIsInstance(w, Failure)
            self.assertEqual(w.failureid, 'python-exception')
            self.assertEqual(w.location.file.givenpath, 'wspy_register.c')
            self.assertEqual(w.location.function.name,
                             'register_all_py_protocols_func')
            self.assertEqual(w.location.line, 159)
            self.assertEqual(w.location.column, 42)
            self.assertTrue(w.customfields['traceback']
                            .startswith('wspy_register.c: In function \'register_all_py_protocols_func\':\n'))

    def test_example_5(self):
        # Ensure that we can load range information from XML
        with open('examples/example-5.xml') as f:
            a = Analysis.from_xml(f)
            self.assertEqual(len(a.results), 1)

            w = a.results[0]
            self.assertIsInstance(w, Issue)
            self.assertEqual(w.location.range_.start.line, 10)
            self.assertEqual(w.location.range_.start.column, 9)
            self.assertEqual(w.location.range_.end.line, 10)
            self.assertEqual(w.location.range_.end.column, 44)

            self.assertEqual(w.location.point, None)

            # The line/column getters use the start:
            self.assertEqual(w.location.line, 10)
            self.assertEqual(w.location.column, 9)

    def test_example_6(self):
        with open('examples/example-6.xml') as f:
            a = Analysis.from_xml(f)
            self.assertEqual(a.metadata.generator.name, 'cpychecker')

            self.assertEqual(len(a.results), 1)
            w = a.results[0]
            self.assertIsInstance(w, Failure)
            self.assertEqual(w.failureid, 'too-complicated')
            self.assertEqual(w.message.text,
                             'this function is too complicated for the'
                             ' reference-count checker to fully analyze:'
                             ' not all paths were analyzed')
            self.assertEqual(w.customfields, None)

    def test_non_ascii_example(self):
        with open('examples/example-non-ascii.xml') as f:
            a = Analysis.from_xml(f)
            self.assertEqual(a.metadata.generator.name, u('\u2620') * 8)

            self.assertEqual(len(a.results), 1)
            w = a.results[0]
            self.assertIsInstance(w, Issue)

            # Verify the Japanese version of
            #  "comparison between signed and unsigned integer expressions"
            # within the message:
            self.assertEqual(w.message.text,
                             (u('\u7b26\u53f7\u4ed8\u304d\u3068\u7b26\u53f7'
                                '\u7121\u3057\u306e\u6574\u6570\u5f0f\u306e'
                                '\u9593\u3067\u306e\u6bd4\u8f03\u3067\u3059')))

            # Verify the "mojibake" Kanji/Hiragana within the notes:
            self.assertIn(u('\u6587\u5b57\u5316\u3051'),
                          w.notes.text)

            self.assertEqual(w.location.function.name, u('oo\u025f'))

    def test_to_xml(self):
        def validate(xmlbytes):
            f = tempfile.NamedTemporaryFile(delete=False)
            f.write(xmlbytes)
            f.flush()
            print("")
            p = subprocess.check_output(['xmllint',
                                         '--relaxng', 'firehose.rng',
                                         '--noout',
                                         f.name])
            # do this by hand: if a test fails, we'll want to inspect the
            # file:
            os.unlink(f.name)

        a, w = self.make_simple_analysis()
        validate(a.to_xml_bytes())

        a, w = self.make_complex_analysis()
        validate(a.to_xml_bytes())

        a, w = self.make_failed_analysis()
        validate(a.to_xml_bytes())

        a, w = self.make_info()
        validate(a.to_xml_bytes())

    def test_xml_roundtrip(self):
        def roundtrip_through_xml(a):
            xmlbytes = a.to_xml_bytes()

            buf = BytesIO(xmlbytes)
            return Analysis.from_xml(buf)

        a1, w = self.make_simple_analysis()
        a2 = roundtrip_through_xml(a1)

        self.assertEqual(a1.metadata, a2.metadata)
        self.assertEqual(a1.results, a2.results)
        self.assertEqual(a1, a2)

        a3, w = self.make_complex_analysis()
        a4 = roundtrip_through_xml(a3)

        self.assertEqual(a3.metadata, a4.metadata)
        self.assertEqual(a3.results, a4.results)
        self.assertEqual(a3, a4)

        a5, f = self.make_failed_analysis()
        a6 = roundtrip_through_xml(a5)

        self.assertEqual(a5.metadata, a6.metadata)
        self.assertEqual(a5.results, a6.results)
        self.assertEqual(a5, a6)

        a7, info = self.make_info()
        a8 = roundtrip_through_xml(a7)

        self.assertEqual(a7.metadata, a8.metadata)
        self.assertEqual(a7.results, a8.results)
        self.assertEqual(a7, a8)

        a9 = Analysis.from_xml('examples/example-non-ascii.xml')
        a10 = roundtrip_through_xml(a9)
        self.assertEqual(a9, a10)

    def test_json_roundtrip(self):
        verbose = False

        def roundtrip_through_json(a):
            jsondict = a.to_json()
            if verbose:
                from pprint import pprint
                pprint(jsondict)
            return Analysis.from_json(jsondict)

        a1, w = self.make_simple_analysis()
        a2 = roundtrip_through_json(a1)

        self.assertEqual(a1.metadata, a2.metadata)
        self.assertEqual(a1.results, a2.results)
        self.assertEqual(a1, a2)

        a3, w = self.make_complex_analysis()
        a4 = roundtrip_through_json(a3)

        self.assertEqual(a3.metadata, a4.metadata)
        self.assertEqual(a3.results, a4.results)
        self.assertEqual(a3, a4)

        a5, f = self.make_failed_analysis()
        a6 = roundtrip_through_json(a5)

        self.assertEqual(a5.metadata, a6.metadata)
        self.assertEqual(a5.results, a6.results)
        self.assertEqual(a5, a6)

        a7, info = self.make_info()
        a8 = roundtrip_through_json(a7)

        self.assertEqual(a7.metadata, a8.metadata)
        self.assertEqual(a7.results, a8.results)
        self.assertEqual(a7, a8)

        a9 = Analysis.from_xml('examples/example-non-ascii.xml')
        a10 = roundtrip_through_json(a9)
        self.assertEqual(a9, a10)

    def test_repr(self):
        # Verify that the various __repr__ methods are sane:
        a, w = self.make_simple_analysis()
        self.assertIn('Analysis(', repr(a))
        self.assertIn('Issue(', repr(a))

        a, w = self.make_complex_analysis()
        self.assertIn('Analysis(', repr(a))
        self.assertIn('Issue(', repr(a))

        a, f = self.make_failed_analysis()
        self.assertIn('Analysis(', repr(a))
        self.assertIn('Failure(', repr(a))

        a, info = self.make_info()
        self.assertIn('Analysis(', repr(a))
        self.assertIn('Info(', repr(a))

    def test_hash(self):
        def compare_hashes(creator):
            a1, w1 = creator()
            a2, w2 = creator()
            self.assertEqual(hash(a1), hash(a2))
        compare_hashes(self.make_simple_analysis)
        compare_hashes(self.make_complex_analysis)
        compare_hashes(self.make_failed_analysis)
        compare_hashes(self.make_info)

    def test_cwe(self):
        # Verify that the CWE methods are sane:
        a, w = self.make_complex_analysis()
        self.assertIsInstance(w.cwe, int)
        self.assertEqual(w.get_cwe_str(), 'CWE-681')
        self.assertEqual(w.get_cwe_url(),
                         'http://cwe.mitre.org/data/definitions/681.html')

        # Verify that they are sane for a warning without a CWE:
        a, w = self.make_simple_analysis()
        self.assertEqual(w.cwe, None)
        self.assertEqual(w.get_cwe_str(), None)
        self.assertEqual(w.get_cwe_url(), None)

    def test_fixup_paths(self):
        # Verify that Report.fixup_files() can make paths absolute:
        a, w = self.make_simple_analysis()

        self.assertEqual(w.location.file.abspath, None)
        a.fixup_files(relativedir='/home/david/coding/test')
        self.assertEqual(w.location.file.abspath, '/home/david/coding/test/foo.c')

    def test_fixup_hashes(self):
        # Verify that Report.fixup_files() can add hashes to files:
        a, w = self.make_simple_analysis()
        w.location.file.givenpath = 'examples/python-src-example.c'
        w.location.file.abspath = None
        self.assertEqual(w.location.file.hash_, None)

        a.fixup_files(hashalg='sha1')
        self.assertEqual(w.location.file.hash_.alg, 'sha1')
        self.assertEqual(w.location.file.hash_.hexdigest,
                         'e978c45fc1779e59d5f8c6c0d534fe2d0a5a7c66')

    def test_gcc_output(self):
        a, w = self.make_simple_analysis()

        output = StringIO()
        w.write_as_gcc_output(output)
        self.assertEqual(output.getvalue(),
                         'foo.c:10:15: warning: something bad involving pointers\n')

        a, w = self.make_complex_analysis()
        output = StringIO()
        w.write_as_gcc_output(output)
        self.assertMultiLineEqual(output.getvalue(),
            ("foo.c: In function 'bar':\n"
             "foo.c:10:15: warning: something bad involving pointers [CWE-681]\n"
             "here is some explanatory text\n"
             "foo.c:7:12: note: first we do this\n"
             "foo.c:8:10: note: then we do that\n"
             "foo.c:10:15: note: then it crashes here\n"))

    def test_debian_source(self):
        """ Test to ensure that Debian source package Sut loading works. """
        with open('examples/example-debian-source.xml') as f:
            a = Analysis.from_xml(f)
            self.assertEqual(a.metadata.generator.name, 'handmade')
            self.assertEqual(a.metadata.generator.version, '0.1')
            self.assertIsInstance(a.metadata.sut, DebianSource)
            self.assertEqual(a.metadata.sut.name, 'python-ethtool')
            self.assertEqual(a.metadata.sut.version, '0.7')
            self.assertEqual(a.metadata.sut.release, '4.1+b1')
            self.assertFalse(hasattr(a.metadata.sut, 'buildarch'))

    def test_debian_binary(self):
        """ Test to ensure that Debian binary package Sut loading works. """
        with open('examples/example-debian-binary.xml') as f:
            a = Analysis.from_xml(f)
            self.assertEqual(a.metadata.generator.name, 'handmade')
            self.assertEqual(a.metadata.generator.version, '0.1')
            self.assertIsInstance(a.metadata.sut, DebianBinary)
            self.assertEqual(a.metadata.sut.name, 'python-ethtool')
            self.assertEqual(a.metadata.sut.version, '0.7')
            self.assertEqual(a.metadata.sut.buildarch, 'amd64')
            self.assertEqual(a.metadata.sut.release, '1.1')

    def parse_xml_bytes(self, xmlbytes):
        f = BytesIO(xmlbytes)
        a = Analysis.from_xml(f)
        f.close()
        return a

    def test_empty_str_field(self):
        a = self.parse_xml_bytes(
            b'''<analysis>
                  <metadata><generator name='test'/></metadata>
                  <results/>
                  <custom-fields>
                     <str-field name="test"/>
                  </custom-fields>
               </analysis>''')
        # Ensure that an empty <str-field> has value '', rather than None:
        self.assertEqual(a.customfields['test'], '')

    def test_set_custom_field(self):
        a, w = self.make_simple_analysis()
        self.assertEqual(a.customfields, None)

        a.set_custom_field('foo', 'bar')
        self.assertNotEqual(a.customfields, None)
        self.assertEqual(a.customfields['foo'], 'bar')



#   Copyright 2013 Red Hat, Inc.
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
#   USA

import os
import unittest

from firehose.model import Issue, Failure

from firehose.parsers.cppcheck import parse_file

class TestParseXml(unittest.TestCase):
    def parse_example(self, filename):
        a = parse_file(os.path.join(os.path.dirname(__file__),
                                    'example-output',
                                    'cppcheck-xml-v2',
                                    filename))
        return a

    def test_example_001(self):
        a = self.parse_example('example-001.xml')
        self.assertEqual(a.metadata.generator.name, 'cppcheck')
        self.assertEqual(a.metadata.generator.version, '1.57')
        self.assertEqual(a.metadata.sut, None)
        self.assertEqual(a.metadata.file_, None)
        self.assertEqual(a.metadata.stats, None)
        self.assertEqual(len(a.results), 7)
        r0 = a.results[0]
        self.assertIsInstance(r0, Issue)
        self.assertEqual(r0.cwe, None)
        self.assertEqual(r0.testid, 'uninitvar')
        self.assertEqual(r0.message.text, 'Uninitialized variable: ret')
        self.assertEqual(r0.notes, None)
        self.assertEqual(r0.location.file.givenpath,
                         'python-ethtool/etherinfo_obj.c')
        self.assertEqual(r0.location.file.abspath, None)
        self.assertEqual(r0.location.function, None)
        self.assertEqual(r0.location.line, 185)
        self.assertEqual(r0.trace, None)
        self.assertEqual(r0.severity, 'error')

    def test_example_002(self):
        a = self.parse_example('example-002.xml')
        self.assertEqual(a.metadata.generator.name, 'cppcheck')
        self.assertEqual(a.metadata.generator.version, '1.58')
        self.assertEqual(a.metadata.sut, None)
        self.assertEqual(a.metadata.file_, None)
        self.assertEqual(a.metadata.stats, None)
        self.assertEqual(len(a.results), 1)
        r0 = a.results[0]
        self.assertIsInstance(r0, Failure)
        self.assertEqual(r0.failureid, 'toomanyconfigs')
        self.assertEqual(r0.location, None)
        self.assertEqual(r0.message.text,
                         ('Too many #ifdef configurations - cppcheck only'
                          ' checks 12 configurations. Use --force to check'
                          ' all configurations. For more details, use'
                          ' --enable=information.'))
        self.assertEqual(r0.customfields['verbose'],
                         ('The checking of the file will be interrupted because'
                          ' there are too many #ifdef configurations. Checking of'
                          ' all #ifdef configurations can be forced by --force'
                          ' command line option or from GUI preferences. However'
                          ' that may increase the checking time. For more details,'
                          ' use --enable=information.'))



#!/usr/bin/env python

#   Copyright 2013 Shuxiong Ye <yeshuxiong@gmail.com>
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#   02110-1301 USA

import re
import sys
import xml.etree.ElementTree as ET

from firehose.model import Message, Function, Point, \
    File, Location, Metadata, Generator, Issue, Analysis

DEBUG=False

# Parser for xml output from findbugs

def parse_file(data_file_obj, findbugs_version=None, sut=None, file_=None,
        stats=None):
    """
    :param data_file_obj:       file object containing findbugs scan result
                                in xml format, it can be generated using 
                                command:
                                fb analyze -xml:withMessages [jar_file]
    :type  data_file_obj:       file object
    :param findbugs_version:    version of findbugs
    :type  findbugs_version:    str

    :return:    Analysis instance
    """
    generator = Generator(name = "findbugs",
            version = findbugs_version)
    metadata = Metadata(generator, sut, file_, stats)
    analysis = Analysis(metadata, [])

    def parse_BugInstance(bugInstance):
        message = Message(bugInstance.find("LongMessage").text)
        # findbugs has no column information
        sourceLine = bugInstance.find("SourceLine")
        point = Point(int(sourceLine.get("start")), 0)
        path = sourceLine.get("sourcepath")
        path = File(path, None)
        method = bugInstance.find("Method")
        if method:
            function = method.find("Message").text
            tmpIndex = function.rfind("In method ") + len("In method ") - 1
            function = Function(function[tmpIndex+1:])
        else:
            function = None
        location = Location(path, function, point)
        if DEBUG:
            print(str(location)+" "+str(message))
        return Issue(None, None, location, message, None, None)

    tree = ET.parse(data_file_obj)
    root = tree.getroot()
    for bugInstance in root.findall("BugInstance"):
        issue=parse_BugInstance(bugInstance)
        if issue:
            analysis.results.append(issue)
        else:
            sys.stderr.write("fail to pass bugInstance=[%s]\n" %
                    str(bugInstance))
    return analysis
 
if __name__ == '__main__':
    if len(sys.argv) != 2:
        sys.stdout.write("Usage: %s [findbugs result files]\n" % 
                sys.argv[0])
    else:
        analysis = parse_file(open(sys.argv[1],"r"))
        sys.stdout.write(str(analysis.to_xml()))
        sys.stdout.write('\n')



#!/usr/bin/env python
#   Copyright 2013 David Malcolm <dmalcolm@redhat.com>
#   Copyright 2013 Red Hat, Inc.
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
#   USA

from distutils.core import setup

setup(name='firehose',
    version='0.5',
    description='Library for working with output from static code analyzers',
    packages=['firehose',
              'firehose.parsers'],
    license='LGPL2.1 or later',
    author='David Malcolm <dmalcolm@redhat.com>',
    url='https://github.com/fedora-static-analysis/firehose',
    classifiers=(
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries',
    )
)



#!/usr/bin/env python

#   Copyright 2013 David Malcolm <dmalcolm@redhat.com>
#   Copyright 2013 Red Hat, Inc.
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
#   USA

import sys
import xml.etree.ElementTree as ET

from firehose.model import Message, Function, Point, \
    File, Location, Generator, Metadata, Analysis, Issue, Notes, Failure, \
    CustomFields

# Parser for output from cppcheck:
#   http://sourceforge.net/apps/mediawiki/cppcheck/index.php?title=Main_Page
# specifically, version 2 of its XML format as generated by:
#   cppcheck PATH_TO_SOURCES --xml --xml-version=2

def parse_file(fileobj, sut=None, file_=None, stats=None):
    tree = ET.parse(fileobj)
    root = tree.getroot()
    node_cppcheck = root.find('cppcheck')
    version = node_cppcheck.get('version')
    node_errors = root.find('errors')

    generator = Generator(name='cppcheck',
                          version=node_cppcheck.get('version'))
    metadata = Metadata(generator, sut, file_, stats)
    analysis = Analysis(metadata, [])

    for node_error in node_errors.findall('error'):
        # e.g.:
        # <error id="nullPointer" severity="error" msg="Possible null pointer dereference: end - otherwise it is redundant to check it against null." verbose="Possible null pointer dereference: end - otherwise it is redundant to check it against null.">
        #  <location file="python-ethtool/ethtool.c" line="139"/>
        #  <location file="python-ethtool/ethtool.c" line="141"/>
        # </error>
        testid = node_error.get('id')
        str_msg = node_error.get('msg')
        str_verbose = node_error.get('verbose')
        message = Message(text=str_msg)
        if str_verbose != str_msg:
            notes = Notes(str_verbose)
        else:
            notes = None

        location_nodes = list(node_error.findall('location'))
        for node_location in location_nodes:
            location=Location(file=File(node_location.get('file'), None),

                              # FIXME: doesn't tell us function name
                              # TODO: can we patch this upstream?
                              function=None,

                              # doesn't emit column
                              point=Point(int(node_location.get('line')), 0)) # FIXME: bogus column
            issue = Issue(None, testid, location, message, notes, None,
                          severity=node_error.get('severity'))
            analysis.results.append(issue)

        if not location_nodes:
            customfields=CustomFields()
            if str_verbose != str_msg:
                customfields['verbose'] = str_verbose
            failure = Failure(failureid=testid,
                              location=None,
                              message=message,
                              customfields=customfields)
            analysis.results.append(failure)

    return analysis

if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("provide a build log file path as the only argument")
    else:
        with open(sys.argv[1]) as data_file:
            analysis = parse_file(data_file)
            sys.stdout.write(str(analysis.to_xml()))
            sys.stdout.write("\n")



#   Copyright 2013 Red Hat, Inc.
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
#   USA

import unittest

import mock

from firehose.parsers import gcc # import parse_warning, parse_file
from firehose.model import Analysis, Issue, Location, File, Point, \
    Function, Message, Sut, Metadata, Generator

FUNC_NAME = 'I am a func name'

FAKE_SUT = Sut()

class TestParseWarning(unittest.TestCase):
    def test_empty(self):
        ret = gcc.parse_warning('', FUNC_NAME)
        self.assertTrue(ret is None)

    def test_new_line(self):
        ret = gcc.parse_warning('\n', FUNC_NAME)
        self.assertTrue(ret is None)

    def test_parse_c(self):
        line = "unix/arlib.c:299:9: warning: ignoring return value of 'fread', declared with attribute warn_unused_result [-Wunused-result]"
        ret = gcc.parse_warning(line, FUNC_NAME)
        self.assertIsInstance(ret, Issue)

    def test_no_metadata(self):
        line = "unix/arlib.c:299:9: warning: ignoring return value of 'fread', declared with attribute warn_unused_result [-Wunused-result]"
        ret = gcc.parse_warning(line, FUNC_NAME)
        self.assertIsInstance(ret, Issue)

    def test_parse_cpp(self):
        line = "num_get_float.cpp:535:29: warning: dereferencing type-punned pointer will break strict-aliasing rules [-Wstrict-aliasing]"
        ret = gcc.parse_warning(line, FUNC_NAME)
        self.assertIsInstance(ret, Issue)

    def test_empty_col(self):
        line = "webkit/glue/webkitclient_impl.cc:411: warning: 'current_mem_usage' may be used uninitialized in this function"
        ret = gcc.parse_warning(line, FUNC_NAME)
        self.assertIsInstance(ret, Issue)
        self.assertEqual(ret.location.column, 0)
        self.assertEqual(ret.location.line, 411)

    def test_values_c(self):
        line = "unix/arlib.c:299:9: warning: ignoring return value of 'fread', declared with attribute warn_unused_result [-Wunused-result]"
        issue = gcc.parse_warning(line, FUNC_NAME)

        # Verify the metadata:
        self.assertEqual(issue.cwe, None)
        self.assertEqual(issue.testid, 'unused-result')
        self.assertIsInstance(issue.location, Location)
        self.assertIsInstance(issue.location.file, File)
        self.assertIsInstance(issue.location.point, Point)
        self.assertIsInstance(issue.location.function, Function)
        self.assertIsInstance(issue.message, Message)
        self.assertEqual(issue.message.text,
            "ignoring return value of 'fread', declared with attribute warn_unused_result")
        self.assertEqual(issue.location.file.givenpath, "unix/arlib.c")
        self.assertEqual(issue.location.file.abspath, None)
        self.assertEqual(issue.location.point.line, 299)
        self.assertEqual(issue.location.point.column, 9)
        self.assertEqual(issue.location.function.name, FUNC_NAME)

    def test_values_cpp(self):
        line = "num_get_float.cpp:535:29: warning: dereferencing type-punned pointer will break strict-aliasing rules [-Wstrict-aliasing]"
        issue = gcc.parse_warning(line, FUNC_NAME)

        self.assertIsInstance(issue.location, Location)
        self.assertIsInstance(issue.location.file, File)
        self.assertIsInstance(issue.location.point, Point)
        self.assertIsInstance(issue.location.function, Function)
        self.assertIsInstance(issue.message, Message)
        self.assertEqual(issue.message.text,
            "dereferencing type-punned pointer will break strict-aliasing rules")
        self.assertEqual(issue.location.file.givenpath, "num_get_float.cpp")
        self.assertEqual(issue.location.file.abspath, None)
        self.assertEqual(issue.location.point.line, 535)
        self.assertEqual(issue.location.point.column, 29)
        self.assertEqual(issue.location.function.name, FUNC_NAME)

    def test_full_path(self):
        line = "/builddir/build/BUILD/libreoffice-3.5.7.2/icc/unxlngi6.pro/misc/build/SampleICC-1.3.2/IccProfLib/IccMpeACS.cpp:203:40: warning: comparison between signed and unsigned integer expressions [-Wsign-compare]"
        issue = gcc.parse_warning(line, FUNC_NAME)

        self.assertEqual(issue.location.file.givenpath,
            "/builddir/build/BUILD/libreoffice-3.5.7.2/icc/unxlngi6.pro/misc/build/SampleICC-1.3.2/IccProfLib/IccMpeACS.cpp")
        self.assertEqual(issue.location.file.abspath, None)

    def test_ignore_zip_warning(self):
        line = "        zip warning: ../../unxlngi6.pro/bin/autotextuser.zip not found or empty"
        issue = gcc.parse_warning(line, FUNC_NAME)
        self.assertTrue(issue is None)

    def test_ignore_java_warning(self):
        line = "    [javac] /builddir/build/BUILD/libreoffice-3.5.7.2/hsqldb/unxlngi6.pro/misc/build/hsqldb/build/build.xml:139: warning: 'includeantruntime' was not set, defaulting to build.sysclasspath=last; set to false for repeatable builds"
        issue = gcc.parse_warning(line, FUNC_NAME)
        self.assertTrue(issue is None)

    def test_parse_no_switch(self):
        line = "unix/arlib.c:299:9: warning: ignoring return value of 'fread', declared with attribute warn_unused_result"
        issue = gcc.parse_warning(line, FUNC_NAME)
        self.assertIsInstance(issue, Issue)


class TestParseFile(unittest.TestCase):
    @staticmethod
    def create_mock_file(lines):
        mock_file = mock.MagicMock()
        mock_file.readlines.return_value = lines
        return mock_file

    @mock.patch.object(gcc, 'parse_warning')
    def test_identify_c(self, mock_parse_warning):
        lines = self.create_mock_file(["unix/arlib.c: In function 'ar_scan':", "ignored by mock"])
        analysis = gcc.parse_file(lines, '4.7.2')
        self.assertEqual(len(analysis.results), 1)
        self.assertEqual(analysis.results[0],
                         mock_parse_warning.return_value)
        self.assertEqual(mock_parse_warning.call_args[0][1], "ar_scan")

    @mock.patch.object(gcc, 'parse_warning')
    def test_identify_cpp(self, mock_parse_warning):
        lines = self.create_mock_file(
            ["/builddir/build/BUILD/libreoffice-3.5.7.2/icc/unxlngi6.pro/misc/build/SampleICC-1.3.2/IccProfLib/IccMpeACS.cpp: In member function 'virtual bool CIccMpeAcs::Read(icUInt32Number, CIccIO*)':",
            "ignored by mock"])
        analysis = gcc.parse_file(lines, '4.7.2')
        self.assertEqual(len(analysis.results), 1)
        self.assertEqual(analysis.results[0],
                         mock_parse_warning.return_value)
        self.assertEqual(mock_parse_warning.call_args[0][1], "virtual bool CIccMpeAcs::Read(icUInt32Number, CIccIO*)")

    @mock.patch.object(gcc, 'parse_warning')
    def test_identify_global(self, mock_parse_warning):
        lines = self.create_mock_file(
            ["/builddir/build/BUILD/libreoffice-3.5.7.2/icc/unxlngi6.pro/misc/build/SampleICC-1.3.2/IccProfLib/IccMpeACS.cpp: At global scope:",
            "ignored by mock"])
        analysis = gcc.parse_file(lines, '4.7.2')
        self.assertEqual(len(analysis.results), 1)
        self.assertEqual(analysis.results[0],
                         mock_parse_warning.return_value)
        self.assertEqual(mock_parse_warning.call_args[0][1], gcc.GLOBAL_FUNC_NAME)

    @mock.patch.object(gcc, 'parse_warning')
    def test_multiple_warnings_per_func(self, mock_parse_warning):
        # we expect that upon reaching "None", it will stop looking for
        # warnings, and thus the last MagicMock won't be counted
        mock_parse_warning.side_effect = [mock.MagicMock(), mock.MagicMock(), None, mock.MagicMock()]
        lines = self.create_mock_file(["unix/arlib.c: In function 'ar_scan':", "", "", "", ""])
        analysis = gcc.parse_file(lines, '4.7.2')
        self.assertEqual(len(analysis.results), 2)
    

# This represents a variety of cases that we should handle.
#
# For the first warning (on line 3), there was no function name given, so do
# we make the function name optional?
"""
In file included from /builddir/build/BUILD/libreoffice-3.5.7.2/icc/unxlngi6.pro/misc/build/SampleICC-1.3.2/IccProfLib/IccMpeACS.h:83:0,
                 from /builddir/build/BUILD/libreoffice-3.5.7.2/icc/unxlngi6.pro/misc/build/SampleICC-1.3.2/IccProfLib/IccMpeACS.cpp:86:
/builddir/build/BUILD/libreoffice-3.5.7.2/icc/unxlngi6.pro/misc/build/SampleICC-1.3.2/IccProfLib/IccTagMPE.h:212:16: warning: unused parameter 'nIterp' [-Wunused-parameter]
/builddir/build/BUILD/libreoffice-3.5.7.2/icc/unxlngi6.pro/misc/build/SampleICC-1.3.2/IccProfLib/IccTagMPE.h:212:16: warning: unused parameter 'pMPE' [-Wunused-parameter]
/builddir/build/BUILD/libreoffice-3.5.7.2/icc/unxlngi6.pro/misc/build/SampleICC-1.3.2/IccProfLib/IccTagMPE.h:213:16: warning: unused parameter 'pDestPixel' [-Wunused-parameter]
/builddir/build/BUILD/libreoffice-3.5.7.2/icc/unxlngi6.pro/misc/build/SampleICC-1.3.2/IccProfLib/IccTagMPE.h:213:16: warning: unused parameter 'pSrcPixel' [-Wunused-parameter]
/builddir/build/BUILD/libreoffice-3.5.7.2/icc/unxlngi6.pro/misc/build/SampleICC-1.3.2/IccProfLib/IccMpeACS.cpp: In member function 'virtual bool CIccMpeAcs::Read(icUInt32Number, CIccIO*)':
/builddir/build/BUILD/libreoffice-3.5.7.2/icc/unxlngi6.pro/misc/build/SampleICC-1.3.2/IccProfLib/IccMpeACS.cpp:203:40: warning: comparison between signed and unsigned integer expressions [-Wsign-compare]
/builddir/build/BUILD/libreoffice-3.5.7.2/icc/unxlngi6.pro/misc/build/SampleICC-1.3.2/IccProfLib/IccMpeACS.cpp: At global scope:
/builddir/build/BUILD/libreoffice-3.5.7.2/icc/unxlngi6.pro/misc/build/SampleICC-1.3.2/IccProfLib/IccMpeACS.cpp:260:6: warning: unused parameter 'nInterp' [-Wunused-parameter]
/builddir/build/BUILD/libreoffice-3.5.7.2/icc/unxlngi6.pro/misc/build/SampleICC-1.3.2/IccProfLib/IccMpeACS.cpp:260:6: warning: unused parameter 'pMPE' [-Wunused-parameter]
"""




#   Copyright 2017 David Malcolm <dmalcolm@redhat.com>
#   Copyright 2017 Red Hat, Inc.
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
#   USA

# Parser for .csv files emitted by "splint"
#   http://splint.org/

from collections import namedtuple
import csv
import os
import re

from firehose.model import Message, Function, Point, Range, \
    File, Location, Generator, Metadata, Analysis, Issue, Sut, Trace, \
    State, Notes, CustomFields

FIELDS = ['warning', 'flag_code', 'flag_name', 'priority',
          'file', 'line', 'column',
          'warning_text', 'additional_text']
WARNING_TEXT_IDX = FIELDS.index('warning_text')

class Row(namedtuple('Row', FIELDS)):
    def to_issue(self):
        """
        Generate an Issue from this csv row.
        """
        location = Location(file=File(givenpath=self.file,
                                      abspath=None),
                            function=None, # FIXME
                            point=Point(int(self.line),
                                        int(self.column)))
        return Issue(cwe=None,
                     testid=self.flag_name,
                     location=location,
                     message=Message(self.warning_text),
                     notes=Notes(self.additional_text),
                     trace=None,
                     severity=self.priority,
                     customfields=None)

def parse_row(row):
    """
    Convert a list generated by csv.reader into a Row
    """
    # splint doesn't escape quotes that occur within the messages, which
    # can lead to extra fields:
    #   https://github.com/ravenexp/splint/issues/6
    # Workaround this by assuming that such quotes occurred in the
    # "Warning Text" field.
    # This fires for warnings 4 and 6
    #
    # This fixes the issue, but there is still some minor information loss
    # e.g. some quote characters in the text disappear, and we sometimes
    # gain a trailing quote character
    if len(row) > len(FIELDS):
        joined_text = ','.join(row[WARNING_TEXT_IDX:-1])
        row = row[:WARNING_TEXT_IDX] + [joined_text, row[-1]]
    return Row(*row)

def parse_splint_csv(path):
    """
    Parse a .csv file written by splint's "-csv FILENAME" option.
    Generate a list of Result instances.
    """
    generator = Generator(name='splint')
    metadata = Metadata(generator, None, None, None)
    analysis = Analysis(metadata, [])
    with open(path, 'r') as f:
        reader = csv.reader(f)
        for raw_row in reader:
            # Skip the initial title row
            if raw_row[0] == 'Warning':
                continue
            rowobj = parse_row(raw_row)
            analysis.results.append(rowobj.to_issue())

    return analysis

def parse_splint_stderr(stderr):
    """
    Parse the stderr from splint (as a string).
    Return a version string, or None.
    """
    lines = stderr.splitlines()
    if not lines:
        return None
    m = re.match("Splint\s+(.*)\s+---.*", lines[0])
    if m:
        return m.group(1)



#!/usr/bin/env python
#
#   Copyright 2017 David Carlos  <ddavidcarlos1392@gmail.com>
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the GNU Lesser General Public
#   License as published by the Free Software Foundation; either
#   version 2.1 of the License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public
#   License along with this library; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301
#   USA

import sys
import re
from subprocess import check_output
from firehose.model import Message, Point, \
    File, Location, Generator, Metadata, Analysis, Issue


def main():
    """ Main entry to flawfinder parser """
    try:
        arg_file = sys.argv[1]
        report = open(arg_file, 'r')
    except IOError as e:
        raise e
    except IndexError:
        print("Missing input file")
    analysis = parse_file(report)
    sys.stdout.write(str(analysis.to_xml()))
    sys.stdout.write('\n')


def parse_file(infile):
    """ Parser flawfinder output

    :infile: file-like object
    :returns: Firehose Analysis object, representing the final XML.

    Flawfinder can generate multiple cwes for a single issue.
    Firehose's models does not supports multiple CWEs.
    For now, when  multiple CWEs ocurrs, we get only the first one.

    A issue was created to track this bug:
    https://github.com/fedora-static-analysis/firehose/issues/35
    """

    line = infile.readline()
    generator = Generator(name='flawfinder',
                          version=get_flawfinder_version(line))
    metadata = Metadata(generator, None, None, None)
    analysis = Analysis(metadata, [])

    # A regex for "filename:linenum:"
    ISSUE_LINE_PATTERN = r"(\S.*)\:([0-9]+)\:"

    # A regex for the reported severity, e.g. "[2]"
    ISSUE_SEVERITY_PATTERN = r"\[([0-9]+)\]"

    # A regex for the reported testid, e.g. "(buffer)"
    ISSUE_TESTID_PATTERN = r"\(([a-z]+)\)"

    WHITESPACE = "\s+"

    FIRST_LINE_PATTERN = (ISSUE_LINE_PATTERN + WHITESPACE +
                     ISSUE_SEVERITY_PATTERN + WHITESPACE +
                     ISSUE_TESTID_PATTERN)
    prog = re.compile(FIRST_LINE_PATTERN)
    while line:
        m = prog.match(line)
        if m:
            issue_path = m.group(1)
            issue_line = m.group(2)
            issue_severity = m.group(3)
            testid = m.group(4)

            location = Location(file=File(issue_path, None),
                                function=None,
                                point=Point(int(issue_line), 0))

            message_line = infile.readline()
            issue_message = ""
            while not prog.search(message_line) and message_line != "\n":
                # Build up issue_message as one line, stripping out
                # extraneous whitespace.
                if issue_message:
                    issue_message += " " + message_line.strip()
                else:
                    issue_message = message_line.strip()
                message_line = infile.readline()

            line = message_line

            cwes = [int(cwe) for cwe in re.findall("CWE-([0-9]+)",
                                                   issue_message)]
            if cwes:
                first_cwe = int(cwes[0])
            else:
                first_cwe = None

            issue = Issue(first_cwe, testid, location,
                          Message(text=issue_message), notes=None,
                          trace=None, severity=issue_severity, customfields=None)

            analysis.results.append(issue)
        else:
            line = infile.readline()

    return analysis


def get_flawfinder_version(first_line):
    """Retrieve flawfinder version from report.
    :first_line: first line of the flawfinder report.
    :returns: flawfinder version.
    """

    pattern = "version\s([0-9]?.[0-9]*)"
    prog = re.compile(pattern)
    try:
        return prog.search(first_line).groups()[0]
    except IndexError:
        return None

if __name__ == '__main__':
    main()
